"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *

from lisflood.hydrological_modules.lakes import *
from lisflood.hydrological_modules.reservoir import *
from lisflood.hydrological_modules.polder import *
from lisflood.hydrological_modules.inflow import *
from lisflood.hydrological_modules.transmission import *
from lisflood.hydrological_modules.kinematic_wave_parallel import kinematicWave, kwpt


class routing(object):

    """
    # ************************************************************
    # ***** ROUTING      *****************************************
    # ************************************************************
    """

    def __init__(self, routing_variable):
        self.var = routing_variable

        self.lakes_module = lakes(self.var)
        self.reservoir_module = reservoir(self.var)
        self.polder_module = polder(self.var)
        self.inflow_module = inflow(self.var)
        self.transmission_module = transmission(self.var)

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
    def initial(self):
        """ initial part of the routing module
        """

        # ------------------------------------------------
        self.var.avgdis = globals.inZero.copy()
        self.var.Beta = loadmap('beta')
        self.var.InvBeta = 1 / self.var.Beta
        # Inverse of beta for kinematic wave
        self.var.ChanLength = loadmap('ChanLength').astype(float)
        self.var.InvChanLength = 1 / self.var.ChanLength
        # Inverse of channel length [1/m]

        self.var.NoRoutSteps = int(np.maximum(1, round(self.var.DtSec / self.var.DtSecChannel,0)))
        # Number of sub-steps based on value of DtSecChannel,
        # or 1 if DtSec is smaller than DtSecChannel

        if option['InitLisflood']:
            self.var.NoRoutSteps = 1
            # InitLisflood is used!
            # so channel routing step is the same as the general time step
        self.var.DtRouting = self.var.DtSec / self.var.NoRoutSteps
        # Corresponding sub-timestep (seconds)
        self.var.InvDtRouting = 1 / self.var.DtRouting
        self.var.InvNoRoutSteps = 1 / float(self.var.NoRoutSteps)
        # inverse for faster calculation inside the dynamic section

# -------------------------- LDD

        self.var.Ldd = lddmask(loadmap('Ldd',pcr=True,lddflag=True), self.var.MaskMap)
        # Cut ldd to size of MaskMap (NEW, 29/9/2004)
        # Prevents 'unsound' ldd if MaskMap covers sub-area of ldd

        # Count (inverse of) upstream area for each pixel
        # Needed if we want to calculate average values of variables
        # upstream of gauge locations

        self.var.UpArea = accuflux(self.var.Ldd, self.var.PixelAreaPcr)
        # Upstream contributing area for each pixel
        # Note that you might expext that values of UpArea would be identical to
        # those of variable CatchArea (see below) at the outflow points.
        # This is NOT actually the case, because outflow points are shifted 1
        # cell in upstream direction in the calculation of CatchArea!
        self.var.InvUpArea = 1 / self.var.UpArea
        # Calculate inverse, so we can multiply in dynamic (faster than divide)

        self.var.IsChannelPcr = pcraster.boolean(loadmap('Channels',pcr=True))
        self.var.IsChannel = np.bool8(compressArray(self.var.IsChannelPcr))

        # Identify channel pixels
        self.var.IsChannelKinematic = self.var.IsChannel.copy()
        # Identify kinematic wave channel pixels
        # (identical to IsChannel, unless dynamic wave is used, see below)
        #self.var.IsStructureKinematic = pcraster.boolean(0)
        self.var.IsStructureKinematic = np.bool8(globals.inZero.copy())

        # Map that identifies special inflow/outflow structures (reservoirs, lakes) within the
        # kinematic wave channel routing. Set to (dummy) value of zero modified in reservoir and lake
        # routines (if those are used)
        LddChan = lddmask(self.var.Ldd, self.var.IsChannelPcr)
        # ldd for Channel network
        self.var.MaskMap = pcraster.boolean(self.var.Ldd)
        # Use boolean version of Ldd as calculation mask
        # (important for correct mass balance check
        # any water generated outside of Ldd won't reach
        # channel anyway)
        self.var.LddToChan = lddrepair(ifthenelse(self.var.IsChannelPcr, 5, self.var.Ldd))
        # Routing of runoff (incl. ground water)en
        AtOutflow = pcraster.boolean(pit(self.var.Ldd))
        # find outlet points...

        if option['dynamicWave']:
            IsChannelDynamic = pcraster.boolean(loadmap('ChannelsDynamic',pcr=True))
            # Identify channel pixels where dynamic wave is used
            self.var.IsChannelKinematic = (self.var.IsChannelPcr == 1) & (IsChannelDynamic == 0)
            # Identify (update) channel pixels where kinematic wave is used
            self.var.LddKinematic = lddmask(self.var.Ldd, self.var.IsChannelKinematic)
            # Ldd for kinematic wave: ends (pit) just before dynamic stretch
            LddDynamic = lddmask(self.var.Ldd, IsChannelDynamic)
            # Ldd for dynamic wave

            # Following statements produce an ldd network that connects the pits in
            # LddKinematic to the nearest downstream dynamic wave pixel
            LddToDyn = lddrepair(ifthenelse(IsChannelDynamic, 5, self.var.Ldd))
            # Temporary ldd: flow paths end in dynamic pixels
            PitsKinematic = cover(pcraster.boolean(pit(self.var.LddKinematic)), 0)
            # Define start of each flow path at pit on LddKinematic
            PathKinToDyn = path(LddToDyn, PitsKinematic)
            # Identify paths that connect pits in LddKinematic to dynamic wave
            # pixels
            LddKinToDyn = lddmask(LddToDyn, PathKinToDyn)
            # Create ldd
            DynWaveBoundaryCondition = boolean(pit(LddDynamic))
            # NEW 12-7-2005 (experimental)
            # Location of boundary condition dynamic wave

            self.var.AtLastPoint = (downstream(self.var.Ldd, AtOutflow) == 1) & (
            AtOutflow != 1) & self.var.IsChannelPcr

            # NEW 23-6-2005
            # Dynamic wave routine gives no outflow out of pits, so we calculate this
            # one cell upstream (WvD)
            # (implies that most downstream cell is not taken into account in mass balance
            # calculations, even if dyn wave is not used)
            # Only include points that are on a channel (otherwise some small 'micro-catchments'
            # are included, for which the mass balance cannot be calculated
            # properly)



        else:
            self.var.LddKinematic = LddChan
            # No dynamic wave, so kinematic ldd equals channel ldd
            self.var.AtLastPoint = AtOutflow
            self.var.AtLastPointC = np.bool8(compressArray(self.var.AtLastPoint))
            # assign unique identifier to each of them


        lddC = compressArray(self.var.LddKinematic)
        inAr = decompress(np.arange(maskinfo['mapC'][0],dtype="int32"))
        # giving a number to each non missing pixel as id
        self.var.downstruct = (compressArray(downstream(self.var.LddKinematic,inAr))).astype("int32")
        # each upstream pixel gets the id of the downstream pixel
        self.var.downstruct[lddC==5] = maskinfo['mapC'][0]
        # all pits gets a high number
        #d3=np.bincount(self.var.down, weights=loadmap('AvgDis'))[:-1]
          # upstream function in numpy

        OutflowPoints = nominal(uniqueid(self.var.AtLastPoint))
            # and assign unique identifier to each of them
        self.var.Catchments = (compressArray(catchment(self.var.Ldd, OutflowPoints))).astype(np.int32)
        CatchArea = np.bincount(self.var.Catchments, weights=self.var.PixelArea)[self.var.Catchments]
        #CatchArea = CatchArea[self.var.Catchments]
        # define catchment for each outflow point
        #CatchArea = areatotal(self.var.PixelArea, self.var.Catchments)

        # Compute area of each catchment [m2]
        # Note: in earlier versions this was calculated using the "areaarea" function,
        # changed to "areatotal" in order to enable handling of grids with spatially
        # variable cell areas (e.g. lat/lon grids)
        self.var.InvCatchArea = 1 / CatchArea
        # inverse of catchment area [1/m2]

# ************************************************************
# ***** CHANNEL GEOMETRY  ************************************
# ************************************************************

        self.var.ChanGrad = np.maximum(loadmap('ChanGrad'), loadmap('ChanGradMin'))
        # avoid calculation of Alpha using ChanGrad=0: this creates MV!
        self.var.CalChanMan = loadmap('CalChanMan')
        self.var.ChanMan = self.var.CalChanMan * loadmap('ChanMan')
        # Manning's n is multiplied by ChanManCal
        # enables calibration for peak timing
        self.var.ChanBottomWidth = loadmap('ChanBottomWidth')
        ChanDepthThreshold = loadmap('ChanDepthThreshold')
        ChanSdXdY = loadmap('ChanSdXdY')
        self.var.ChanUpperWidth = self.var.ChanBottomWidth + 2 * ChanSdXdY * ChanDepthThreshold
        # Channel upper width [m]
        self.var.TotalCrossSectionAreaBankFull = 0.5 * \
            ChanDepthThreshold * (self.var.ChanUpperWidth + self.var.ChanBottomWidth)
        # Area (sq m) of bank full discharge cross section [m2]
        # (trapezoid area equation)
        TotalCrossSectionAreaHalfBankFull = 0.5 * self.var.TotalCrossSectionAreaBankFull
        # Cross-sectional area at half bankfull [m2]
        # This can be used to initialise channel flow (see below)


        TotalCrossSectionAreaInitValue = loadmap('TotalCrossSectionAreaInitValue')
        self.var.TotalCrossSectionArea = np.where(TotalCrossSectionAreaInitValue == -9999, TotalCrossSectionAreaHalfBankFull, TotalCrossSectionAreaInitValue)
        # Total cross-sectional area [m2]: if initial value in binding equals -9999 the value at half bankfull is used,
        # otherwise TotalCrossSectionAreaInitValue (typically end map from previous simulation)

        if option['SplitRouting']:
            CrossSection2AreaInitValue = loadmap('CrossSection2AreaInitValue')
            self.var.CrossSection2Area = np.where(CrossSection2AreaInitValue == -9999, globals.inZero, CrossSection2AreaInitValue)
            # cross-sectional area [m2] for 2nd line of routing: if initial value in binding equals -9999 the value is set to 0
            # otherwise CrossSection2AreaInitValue (typically end map from previous simulation)

            PrevSideflowInitValue = loadmap('PrevSideflowInitValue')


            self.var.Sideflow1Chan = np.where(PrevSideflowInitValue == -9999, globals.inZero, PrevSideflowInitValue)
            # sideflow from previous run for 1st line of routing: if initial value in binding equals -9999 the value is set to 0
            # otherwise PrevSideflowInitValue (typically end map from previous simulation)

# ************************************************************
# ***** CHANNEL ALPHA (KIN. WAVE)*****************************
# ************************************************************
# Following calculations are needed to calculate Alpha parameter in kinematic
# wave. Alpha currently fixed at half of bankful depth (this may change in
# future versions!)

        ChanWaterDepthAlpha = np.where(self.var.IsChannel, 0.5 * ChanDepthThreshold, 0.0)
        # Reference water depth for calculation of Alpha: half of bankfull
        self.var.ChanWettedPerimeterAlpha = self.var.ChanBottomWidth + 2 * \
            np.sqrt(np.square(ChanWaterDepthAlpha) + np.square(ChanWaterDepthAlpha * ChanSdXdY))
        # Channel wetted perimeter [m](Pythagoras)
        AlpTermChan = (self.var.ChanMan / (np.sqrt(self.var.ChanGrad))) ** self.var.Beta
        self.var.AlpPow = 2.0 / 3.0 * self.var.Beta

        self.var.ChannelAlpha = (AlpTermChan * (self.var.ChanWettedPerimeterAlpha ** self.var.AlpPow)).astype(float)
        self.var.InvChannelAlpha = 1 / self.var.ChannelAlpha
        # ChannelAlpha for kinematic wave

# ************************************************************
# ***** CHANNEL INITIAL DISCHARGE ****************************
# ************************************************************

        self.var.ChanM3 = self.var.TotalCrossSectionArea * self.var.ChanLength
        # channel water volume [m3]
        self.var.ChanIniM3 = self.var.ChanM3.copy()
        self.var.ChanM3Kin = self.var.ChanIniM3.copy().astype(float)
        # Initialise water volume in kinematic wave channels [m3]
        self.var.ChanQKin = np.where(self.var.ChannelAlpha > 0,
             (self.var.TotalCrossSectionArea / self.var.ChannelAlpha) ** self.var.InvBeta, 0).astype(float)


        # Initialise discharge at kinematic wave pixels (note that InvBeta is
        # simply 1/beta, computational efficiency!)

        self.var.CumQ = globals.inZero.copy()
        # ininialise sum of discharge to calculate average

# ************************************************************
# ***** CHANNEL INITIAL DYNAMIC WAVE *************************
# ************************************************************
        if option['dynamicWave']:
            dummy=0
            # TO DO !!!!!!!!!!!!!!!!!!!!

       #     lookchan = lookupstate(TabCrossSections, ChanCrossSections, ChanBottomLevel, self.var.ChanLength,
       #                            DynWaveConstantHeadBoundary + ChanBottomLevel)
       #     ChanIniM3 = ifthenelse(AtOutflow, lookchan, ChanIniM3)
            # Correct ChanIniM3 for constant head boundary in pit (only if
            # dynamic wave is used)
       #     ChanM3Dyn = ChanIniM3
            # Set volume of water in dynamic wave channel to initial value
            # (note that initial condition is expressed as a state in [m3] for the dynamic wave,
            # and as a rate [m3/s] for the kinematic wave (a bit confusing)

            # Estimate number of iterations needed in first time step (based on Courant criterium)
            # TO DO !!!!!!!!!!!!!!!!!!!!
        #    Potential = lookuppotential(
        #        TabCrossSections, ChanCrossSections, ChanBottomLevel, self.var.ChanLength, ChanM3Dyn)
            # Potential
        #    WaterLevelDyn = Potential - ChanBottomLevel
            # Water level [m above bottom level)
        #    WaveCelerityDyn = pcraster.sqrt(9.81 * WaterLevelDyn)
            # Dynamic wave celerity [m/s]
        #    CourantDynamic = self.var.DtSec * \
        #        (WaveCelerityDyn + 2) / self.var.ChanLength
            # Courant number for dynamic wave
            # We don't know the water velocity at this time so
            # we just guess it's 2 m/s (Odra tests show that flow velocity
            # is typically much lower than wave celerity, and 2 m/s is quite
            # high already so this gives a pretty conservative/safe estimate
            # for DynWaveIterations)
        #    DynWaveIterationsTemp = max(
        #        1, roundup(CourantDynamic / CourantDynamicCrit))
        #    DynWaveIterations = ordinal(mapmaximum(DynWaveIterationsTemp))
            # Number of sub-steps needed for required numerical
            # accuracy. Always greater than or equal to 1
            # (otherwise division by zero!)

            # TEST
            # If polder option is used, we need an estimate of the initial channel discharge, but we don't know this
            # for the dynamic wave pixels (since only initial state is known)! Try if this works (dyn wave flux based on zero inflow 1 iteration)
            # Note that resulting ChanQ is ONLY used in the polder routine!!!
            # Since we need instantaneous estimate at start of time step, a
            # ChanQM3Dyn is calculated for one single one-second time step!!!

        #    ChanQDyn = dynwaveflux(TabCrossSections,
        #                           ChanCrossSections,
        #                           LddDynamic,
        #                           ChanIniM3,
        #                           0.0,
        #                           ChanBottomLevel,
        #                           self.var.ChanMan,
        #                           self.var.ChanLength,
        #                           1,
        #                           1,
        #                           DynWaveBoundaryCondition)
            # Compute volume and discharge in channel after dynamic wave
            # ChanM3Dyn in [cu m]
            # ChanQDyn in [cu m / s]
        #    self.var.ChanQ = ifthenelse(
        #        IsChannelDynamic, ChanQDyn, self.var.ChanQKin)
            # Channel discharge: combine results of kinematic and dynamic wave
        else:


            # ***** NO DYNAMIC WAVE *************************
            # Dummy code if dynamic wave is not used, in which case ChanQ equals ChanQKin
            # (needed only for polder routine)

            PrevDischarge = loadmap('PrevDischarge')
            self.var.ChanQ = np.where(PrevDischarge == -9999, self.var.ChanQKin, PrevDischarge)
            # initialise channel discharge: cold start: equal to ChanQKin
            # [m3/s]

        # Initialising cumulative output variables
        # These are all needed to compute the cumulative mass balance error

        self.var.DischargeM3Out = globals.inZero.copy()
        # cumulative discharge at outlet [m3]
        self.var.TotalQInM3 = globals.inZero.copy()
        # cumulative inflow from inflow hydrographs [m3]

        #self.var.sumDis = globals.inZero.copy()
        self.var.sumDis = globals.inZero.copy()
        self.var.sumIn = globals.inZero.copy()


# --------------------------------------------------------------------------

    def initialSecond(self):
        """ initial part of the second channel routing module
        """
        self.var.ChannelAlpha2 = None # default value, if split-routing is not active and only water is routed only in the main channel
        # ************************************************************
        # ***** CHANNEL INITIAL SPLIT UP IN SECOND CHANNEL************
        # ************************************************************
        if option['SplitRouting']:

            ChanMan2 = (self.var.ChanMan / self.var.CalChanMan) * loadmap('CalChanMan2')
            AlpTermChan2 = (ChanMan2 / (np.sqrt(self.var.ChanGrad))) ** self.var.Beta
            self.var.ChannelAlpha2 = (AlpTermChan2 * (self.var.ChanWettedPerimeterAlpha ** self.var.AlpPow)).astype(float)
            self.var.InvChannelAlpha2 = 1 / self.var.ChannelAlpha2
            # calculating second Alpha for second (virtual) channel

            if not(option['InitLisflood']):

                self.var.QLimit = loadmap('AvgDis') * loadmap('QSplitMult')
                self.var.M3Limit = self.var.ChannelAlpha * self.var.ChanLength * (self.var.QLimit ** self.var.Beta)
                # Over bankful discharge starts at QLimit
                # lower discharge limit for second line of routing
                # set to mutiple of average discharge (map from prerun)
                # QSplitMult =2 is around 90 to 95% of Q

                ###############################################
                # CM mod
                # QLimit should NOT be dependent on the NoRoutSteps (number of routing steps)
                # self.var.QLimit = self.var.QLimit / self.var.NoRoutSteps #original

                # TEMPORARY WORKAROUND FOR EFAS XDOM!!!!!!!!!!
                # This must be removed
                self.var.QLimit = self.var.QLimit / 24.0
                ###############################################

                self.var.Chan2M3Start = self.var.ChannelAlpha2 * self.var.ChanLength * (self.var.QLimit ** self.var.Beta)
                # virtual amount of water in the channel through second line
                self.var.Chan2QStart = self.var.QLimit - compressArray(upstream(self.var.LddKinematic, decompress(self.var.QLimit)))
                # because kinematic routing with a low amount of discharge leads to long travel time:
                # Starting Q for second line is set to a higher value
                self.var.Chan2M3Kin = np.maximum(self.var.CrossSection2Area * self.var.ChanLength + self.var.Chan2M3Start, 0) # TEMPORARY SOLUTION TO VIRTUAL WATER PROBLEM
                self.var.ChanM3Kin = np.maximum(self.var.ChanM3 - self.var.Chan2M3Kin + self.var.Chan2M3Start, 0) # TEMPORARY SOLUTION TO VIRTUAL WATER PROBLEM
                self.var.Chan2QKin = (self.var.Chan2M3Kin*self.var.InvChanLength*self.var.InvChannelAlpha2)**(self.var.InvBeta)
                self.var.ChanQKin = (self.var.ChanM3Kin*self.var.InvChanLength*self.var.InvChannelAlpha)**(self.var.InvBeta)

        # Initialise parallel kinematic wave router: main channel-only routing if self.var.ChannelAlpha2 is None; else split-routing(main channel + floodplains)
        self.river_router = kinematicWave(compressArray(self.var.LddKinematic), ~maskinfo["mask"], self.var.ChannelAlpha,\
                                          self.var.Beta, self.var.ChanLength, self.var.DtRouting,\
                                          int(binding["numCPUs_parallelKinematicWave"]), alpha_floodplains=self.var.ChannelAlpha2)

               

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------



    def dynamic(self, NoRoutingExecuted):
        """ dynamic part of the routing subtime module
        """

        if not(option['InitLisflood']):    # only with no InitLisflood
            self.lakes_module.dynamic_inloop(NoRoutingExecuted)
            self.reservoir_module.dynamic_inloop(NoRoutingExecuted)
            self.polder_module.dynamic_inloop()

        # End only with no Lisflood (no reservoirs, lakes and polder with
        # initLisflood)

        self.inflow_module.dynamic_inloop(NoRoutingExecuted)
        self.transmission_module.dynamic_inloop()

        # ************************************************************
        # ***** CHANNEL FLOW ROUTING: KINEMATIC WAVE  ****************
        # ************************************************************



        if not(option['dynamicWave']):

            # ************************************************************
            # ***** SIDEFLOW
            # ************************************************************



            SideflowChanM3 = self.var.ToChanM3RunoffDt.copy()

            if option['openwaterevapo']:
                SideflowChanM3 -= self.var.EvaAddM3Dt
            if option['wateruse']:
                SideflowChanM3 -= self.var.WUseAddM3Dt
            if option['inflow']:
                SideflowChanM3 += self.var.QInDt
            if option['TransLoss']:
                SideflowChanM3 -= self.var.TransLossM3Dt
            if not(option['InitLisflood']):    # only with no InitLisflood
                if option['simulateLakes']:
                    SideflowChanM3 += self.var.QLakeOutM3Dt
                if option['simulateReservoirs']:
                    SideflowChanM3 += self.var.QResOutM3Dt
                if option['simulatePolders']:
                    SideflowChanM3 -= self.var.ChannelToPolderM3Dt

            # Runoff (surface runoff + flow out of Upper and Lower Zone), outflow from
            # reservoirs and lakes and inflow from external hydrographs are added to the channel
            # system (here in [cu m])
            #
            # NOTE: polders currently don't work with kinematic wave, but nevertheless
            # ChannelToPolderM3 is already included in sideflow term (so it's there in case
            # the polder routine is ever modified to make it work with kin. wave)
            # Because of wateruse Sideflow might get even smaller than 0,
            # instead of inflow its outflow

            SideflowChan = np.where(self.var.IsChannelKinematic, SideflowChanM3 * self.var.InvChanLength * self.var.InvDtRouting,0)
            # SideflowChan=if(IsChannelKinematic, SideflowChanM3*InvChanLength*InvDtRouting);
            # Sideflow expressed in [cu m /s / m channel length]



            # ************************************************************
            # ***** KINEMATIC WAVE                        ****************
            # ************************************************************

            if option['InitLisflood'] or (not(option['SplitRouting'])):
                # if InitLisflood no split routing is use
                #  ---- Single Routing ---------------
                # No split routing
                # side flow consists of runoff (incl. groundwater), inflow from reservoirs (optional) and external inflow hydrographs (optional)
                SideflowChan[np.isnan(SideflowChan)] = 0 # TEMPORARY FIX - SEE DEBUG ABOVE!

                # ChanQKin in [cu m / s]
                self.river_router.kinematicWaveRouting(self.var.ChanQKin, SideflowChan, "main_channel")
                self.var.ChanM3Kin = self.var.ChanLength * self.var.ChannelAlpha * self.var.ChanQKin**self.var.Beta
                # Volume in channel at end of computation step
                #self.var.ChanQKin=pcraster.max(self.var.ChanQKin,0)
                self.var.ChanQ=np.maximum(self.var.ChanQKin,0)
                # at single kin. ChanQ is the same
                self.var.sumDisDay+=self.var.ChanQ
                # Total channel storage [cu m], equal to ChanM3Kin
                #self.var.ChanQ = maxpcr(self.var.ChanQKin, null)

            else:
                #  ---- Double Routing ---------------
                # routing is split in two (virtual) channels)

                # Ad
                SideflowRatio=np.where((self.var.ChanM3Kin + self.var.Chan2M3Kin) > 0,self.var.ChanM3Kin/(self.var.ChanM3Kin+self.var.Chan2M3Kin),0.0)

                # CM ##################################
                self.var.Sideflow1Chan = np.where(self.var.ChanM3Kin > self.var.M3Limit, SideflowRatio*SideflowChan, SideflowChan)
                # This is creating instability because ChanM3Kin can be < M3Limit between two routing sub-steps
                #######################################
                # self.var.Sideflow1Chan = np.where((self.var.ChanM3Kin + self.var.Chan2M3Kin-self.var.Chan2M3Start) > self.var.M3Limit,
                #                                   SideflowRatio*SideflowChan, SideflowChan)

                self.var.Sideflow1Chan = np.where(np.abs(SideflowChan) < 1e-7, SideflowChan, self.var.Sideflow1Chan)
                # too small values are avoided
                Sideflow2Chan = SideflowChan - self.var.Sideflow1Chan
               
                Sideflow2Chan = Sideflow2Chan + self.var.Chan2QStart * self.var.InvChanLength   # originale
                # as kinematic wave gets slower with less water
                # a constant amount of water has to be added
                # -> add QLimit discharge

                # Main channel routing
                self.river_router.kinematicWaveRouting(self.var.ChanQKin, self.var.Sideflow1Chan, "main_channel")
                self.var.ChanM3Kin = self.var.ChanLength * self.var.ChannelAlpha * self.var.ChanQKin**self.var.Beta

                # Floodplains routing
                self.river_router.kinematicWaveRouting(self.var.Chan2QKin, Sideflow2Chan, "floodplains")
                self.var.Chan2M3Kin = self.var.ChanLength * self.var.ChannelAlpha2 * self.var.Chan2QKin**self.var.Beta
                self.var.CrossSection2Area = (self.var.Chan2M3Kin - self.var.Chan2M3Start) * self.var.InvChanLength # wet cross-section area of floodplain
                self.var.ChanQ = np.maximum(self.var.ChanQKin + self.var.Chan2QKin - self.var.QLimit, 0)
                # Main channel routing and floodplains routing

                # ChanQ=max(ChanQKin+Chan2QKin-QLimit,0.0);
                # Channel discharge: equal to ChanQKin [cu m / s]
                # End splitrouting
                self.var.sumDisDay+=self.var.ChanQ
                # ----------End splitrouting-------------------------------------------------


            
            TotalCrossSectionArea = np.maximum(self.var.ChanM3Kin*self.var.InvChanLength,0.01)

            self.var.FlowVelocity = np.minimum(self.var.ChanQKin/TotalCrossSectionArea, 0.36*self.var.ChanQKin**0.24)
              # Channel velocity (m/s); dividing Q (m3/s) by CrossSectionArea (m2)
              # avoid extreme velocities by using the Wollheim 2006 equation
              # assume 0.1 for upstream areas (outside ChanLdd)
            self.var.FlowVelocity *= np.minimum(np.sqrt(self.var.PixelArea)*self.var.InvChanLength,1);
	          # reduction for sinuosity of channels
            self.var.TravelDistance=self.var.FlowVelocity*self.var.DtSec;
	          # if flow is fast, Traveltime=1, TravelDistance is high: Pixellength*DtSec
	          # if flow is slow, Traveltime=DtSec then TravelDistance=PixelLength
	          # maximum set to 30km/day for 5km cell, is at DtSec/Traveltime=6, is at Traveltime<DtSec/6



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""


from lisflood.global_modules.add1 import *
from kinematic_wave_parallel import kinematicWave


class surface_routing(object):

    """
    # ************************************************************
    # ***** SURFACE ROUTING **************************************
    # ************************************************************
    """

    def __init__(self, surface_routing_variable):
        self.var = surface_routing_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the surface_routing module
        """
        # CM mod
        OFM3OtherInit = loadmap('OFOtherInitValue')
        OFM3ForestInit = loadmap('OFForestInitValue')
        OFM3DirectInit = loadmap('OFDirectInitValue')
        self.var.WaterDepth = globals.inZero.copy()

        # self.var.WaterDepthInit =loadmap('WaterDepthInitValue')
        # self.var.WaterDepthInit = makenumpy(self.var.WaterDepthInit)
        # self.var.WaterDepth = self.var.WaterDepthInit.copy()
        # initial overland flow water depth [mm]
        # for initial water in CHANNEL see CHANNEL GEOMETRY section below!
        ## end CM mod

        self.var.OFM3Other = makenumpy(OFM3OtherInit)
        self.var.OFM3Forest = makenumpy(OFM3ForestInit)
        self.var.OFM3Direct = makenumpy(OFM3DirectInit)

# ************************************************************
# ***** ROUTING OF SURFACE RUNOFF
# ************************************************************

        Grad = np.maximum(loadmap('Grad'), loadmap('GradMin'))
        # Set gradient to minimum value to prevent MV creation later on.

        self.var.NoSubStepsOF = 1
        # Number of sub-steps applied in kinematic wave routing of the Overland Flow.
        # Number of sub-steps applied in kinematic wave routing. Currently fixed
        # at 1 (number of timeslices in kinwavestate, flux must have ordinal datatype!)

        OFWettedPerimeter = self.var.PixelLength + 2 * self.var.MMtoM * loadmap('OFDepRef')
        # Wetted perimeter overland flow [m] pixel width +
        # 2 times fixed reference depth
        # (Note that using grid size as flow width is a bit odd, as results will depend on cell size!)

        AlpTermOF = [(self.var.NManning[0] / (np.sqrt(Grad))) ** self.var.Beta, (self.var.NManning[1] /
                (np.sqrt(Grad))) ** self.var.Beta, (self.var.NManning[2] / (np.sqrt(Grad))) ** self.var.Beta]
        # self.var.OFAlpha = [AlpTermOF[0]*(OFWettedPerimeter**AlpPow),AlpTermOF[1]*(OFWettedPerimeter**AlpPow),AlpTermOF[2]*(OFWettedPerimeter**AlpPow)]

        self.var.OFAlphaOther = (AlpTermOF[0] * (OFWettedPerimeter ** self.var.AlpPow)).astype(float)
        self.var.OFAlphaForest = (AlpTermOF[1] * (OFWettedPerimeter ** self.var.AlpPow)).astype(float)
        self.var.OFAlphaDirect = (AlpTermOF[2] * (OFWettedPerimeter ** self.var.AlpPow)).astype(float)
        self.var.InvOFAlphaOther  = 1 / self.var.OFAlphaOther
        self.var.InvOFAlphaForest = 1 / self.var.OFAlphaForest
        self.var.InvOFAlphaDirect = 1 / self.var.OFAlphaDirect

        # Alpha to separate int 3 different overland routing: forest, water and sealed area, remaining area
        # overland flow Alpha for kinematic wave
        OFCrossSectionArea = self.var.MMtoM * self.var.WaterDepth * self.var.PixelLength
        # Overland flow initial cross-sectional area [m2]
        OFM3all = OFCrossSectionArea * self.var.PixelLength
        # Initial overland flow storage [m3]
        # self.var.OFM3=[cover(OFM3all*self.var.OtherFraction,scalar(0.0)),cover(OFM3all*self.var.ForestFraction,scalar(0.0)),cover(OFM3all*(self.var.DirectRunoffFraction+self.var.WaterFraction),scalar(0.0))]


        # Initial overland discharge [m3 s-1]
        self.var.OFQDirect = ((self.var.OFM3Direct * self.var.InvPixelLength * self.var.InvOFAlphaDirect)**(self.var.InvBeta)).astype(float)
        self.var.OFQOther = ((self.var.OFM3Other * self.var.InvPixelLength * self.var.InvOFAlphaOther)**(self.var.InvBeta)).astype(float)
        self.var.OFQForest = ((self.var.OFM3Forest * self.var.InvPixelLength * self.var.InvOFAlphaForest)**(self.var.InvBeta)).astype(float)

    def initialSecond(self):
        """ 2nd initialisation part of the surface routing module:
            to be called after all needed parameters are set (PixelLength, LddToChan, Alpha...)
            PRE-PROCESSING OF FLOW DIRECTION MATRIX FOR PARALLELISED KINEMATIC WAVE ROUTING
        """
        dt_surf_routing = self.var.DtSec / self.var.NoSubStepsOF
        land_mask = ~maskinfo["mask"]
        num_threads = int(binding["numCPUs_parallelKinematicWave"])
        self.direct_surface_router = kinematicWave(compressArray(self.var.LddToChan), land_mask, self.var.OFAlphaDirect, self.var.Beta,\
                                                   self.var.PixelLength, dt_surf_routing, num_threads)
        self.other_surface_router = kinematicWave(compressArray(self.var.LddToChan), land_mask, self.var.OFAlphaOther, self.var.Beta,\
                                                  self.var.PixelLength, dt_surf_routing, num_threads)
        self.forest_surface_router = kinematicWave(compressArray(self.var.LddToChan), land_mask, self.var.OFAlphaForest, self.var.Beta,\
                                                   self.var.PixelLength, dt_surf_routing, num_threads)


    def dynamic(self):
        """ dynamic part of the surface routing module
        """
        # ************************************************************
        # ***** COMPONENTS OF RUNOFF                               ***
        # ************************************************************
        self.var.SurfaceRunOther = self.var.OtherFraction * \
            np.maximum(self.var.AvailableWaterForInfiltration[0] - self.var.Infiltration[0], 0)
        self.var.SurfaceRunForest = self.var.ForestFraction * \
            np.maximum(self.var.AvailableWaterForInfiltration[1] - self.var.Infiltration[1], 0)
        self.var.SurfaceRunIrrigation = self.var.IrrigationFraction * \
            np.maximum(self.var.AvailableWaterForInfiltration[2] - self.var.Infiltration[2], 0)


        self.var.SurfaceRunoff = self.var.DirectRunoff + self.var.SurfaceRunOther + self.var.SurfaceRunForest + self.var.SurfaceRunIrrigation
        # Surface runoff for this time step (mm)
        # Note that SurfaceRunoff ONLY includes surface runoff generated during current time
        # step (WaterDepth includes routed runoff from previous steps as well)

        self.var.TotalRunoff = self.var.SurfaceRunoff + self.var.UZOutflowPixel + self.var.LZOutflowToChannelPixel

        # ************************************************************
        # ***** ROUTING SURFACE RUNOFF TO CHANNEL ********************
        # ************************************************************

        # Domain: whole pixel
        # Routing of overland flow to channel using kinematic wave
        # Note that all 'new' water is added as side-flow
        SideflowDirect = self.var.DirectRunoff * self.var.MMtoM3 * self.var.InvPixelLength * self.var.InvDtSec
        SideflowOther = (self.var.SurfaceRunOther + self.var.SurfaceRunIrrigation) * self.var.MMtoM3 * self.var.InvPixelLength * self.var.InvDtSec
        SideflowForest = self.var.SurfaceRunForest * self.var.MMtoM3 * self.var.InvPixelLength * self.var.InvDtSec
        # All surface runoff that is generated during current time step added as side flow [m3/s/m pixel-length]

        self.direct_surface_router.kinematicWaveRouting(self.var.OFQDirect, SideflowDirect)
        self.other_surface_router.kinematicWaveRouting(self.var.OFQOther, SideflowOther)
        self.forest_surface_router.kinematicWaveRouting(self.var.OFQForest, SideflowForest)

# to PCRASTER

        #SideflowDirect =  decompress(self.var.DirectRunoff * self.var.MMtoM3 * self.var.InvPixelLength * self.var.InvDtSec)
        #SideflowOther =  decompress((self.var.SurfaceRunOther + self.var.SurfaceRunIrrigation) * self.var.MMtoM3 * self.var.InvPixelLength * self.var.InvDtSec)
        #SideflowForest =  decompress(self.var.SurfaceRunForest * self.var.MMtoM3 * self.var.InvPixelLength * self.var.InvDtSec)
        # All surface runoff that is generated during current time
        # step added as side flow [m3/s/m pixel-length]

        # OFM3,OFQ=kinwavestate,kinwaveflux(LddToChan,OFM3,SideflowOF,OFAlpha,beta,NoSubStepsOF,DtSec,PixelLength)
        # self.var.OFQ=kinwaveflux(self.var.LddToChan,self.var.OFM3,SideflowOF,self.var.OFAlpha,self.var.Beta,self.var.NoSubStepsOF,self.var.DtSec,self.var.PixelLength)
        # self.var.OFM3=kinwavestate(self.var.LddToChan,self.var.OFM3,SideflowOF,self.var.OFAlpha,self.var.Beta,self.var.NoSubStepsOF,self.var.DtSec,self.var.PixelLength)

        #pcrOFM3Direct = decompress(self.var.OFM3Direct)
        #pcrOFM3Other = decompress(self.var.OFM3Other)
        #pcrOFM3Forest = decompress(self.var.OFM3Forest)

    #From here in PCRASTER format!

        #self.var.OFM3Direct = compressArray(kinwavestate(self.var.LddToChan, pcrOFM3Direct, SideflowDirect,
        #                      self.var.OFAlphaDirect, self.var.Beta, self.var.NoSubStepsOF, self.var.DtSec, self.var.PixelLengthPcr))
        #self.var.OFM3Other = compressArray(kinwavestate(self.var.LddToChan, pcrOFM3Other, SideflowOther,
        #                      self.var.OFAlphaOther, self.var.Beta, self.var.NoSubStepsOF, self.var.DtSec, self.var.PixelLengthPcr))
        #self.var.OFM3Forest = compressArray(kinwavestate(self.var.LddToChan, pcrOFM3Forest, SideflowForest,
        #                      self.var.OFAlphaForest, self.var.Beta, self.var.NoSubStepsOF, self.var.DtSec, self.var.PixelLengthPcr))
        # Route overland flow to channel using kinematic wave
        # OFQ in [m3/s]
    # End PCRASTER

        #self.var.OFQDirect = (self.var.OFM3Direct * self.var.InvPixelLength *
        #                    self.var.InvOFAlphaDirect)**(self.var.InvBeta)
        #self.var.OFQOther =(self.var.OFM3Other * self.var.InvPixelLength *
        #                    self.var.InvOFAlphaOther)**(self.var.InvBeta)
        #self.var.OFQForest=(self.var.OFM3Forest * self.var.InvPixelLength *
        #                    self.var.InvOFAlphaForest)**(self.var.InvBeta)

        self.var.OFM3Direct = self.var.PixelLength * self.var.OFAlphaDirect * self.var.OFQDirect**self.var.Beta
        self.var.OFM3Other = self.var.PixelLength * self.var.OFAlphaOther * self.var.OFQOther**self.var.Beta
        self.var.OFM3Forest = self.var.PixelLength * self.var.OFAlphaForest * self.var.OFQForest**self.var.Beta
        self.var.Qall = self.var.OFQDirect + self.var.OFQOther + self.var.OFQForest
        self.var.M3all = self.var.OFM3Direct + self.var.OFM3Other + self.var.OFM3Forest
        # Overland flow storage [m3]

        self.var.OFToChanM3 = np.where(self.var.IsChannel, self.var.Qall * self.var.DtSec, 0)

        # Overland flow in channel pixels (in [m3])is added to channel

        self.var.WaterDepth = self.var.M3all * self.var.M3toMM
        # Update water depth [mm]

        ## self.var.ToChanM3Runoff = accuflux(self.var.LddToChan, (decompress(self.var.UZOutflowPixel) + decompress(self.var.LZOutflowToChannelPixel)) * self.var.MMtoM3) + self.var.OFToChanM3

        # All runoff that enters the channel: groundwater + surface runoff
        # Note that all groundwater/inter-flow is routed to nearest river pixel
        # within one time step
        self.var.ToChanM3Runoff =  (self.var.UZOutflowPixel + self.var.LZOutflowToChannelPixel) * self.var.MMtoM3 + self.var.OFToChanM3
        self.var.ToChanM3RunoffDt = self.var.ToChanM3Runoff * self.var.InvNoRoutSteps
        # runoff of 1 substep



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *


class snow(object):

    """
    # ************************************************************
    # ***** RAIN AND SNOW *****************************************
    # ************************************************************

    # Domain: snow calculations evaluated for center points of 3 sub-pixel
    # snow zones A, B, and C, which each occupy one-third of the pixel surface
    #
    # Variables 'snow' and 'rain' at end of this module are the pixel-average snowfall and rain
    #
    # Zone A: lower third
    # Zone B: center third
    # Zone C: upper third
    """

    def __init__(self, snow_variable):
        self.var = snow_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the snow module
        """
        
        self.var.DeltaTSnow = 0.9674 * loadmap('ElevationStD') * loadmap('TemperatureLapseRate')

        # Difference between (average) air temperature at average elevation of
        # pixel and centers of upper- and lower elevation zones [deg C]
        # ElevationStD:   Standard Deviation of the DEM from Bodis (2009)
        # 0.9674:    Quantile of the normal distribution: u(0,833)=0.9674
        #              to split the pixel in 3 equal parts.
        self.var.SnowDayDegrees = 360 / 365.25
        # day of the year to degrees: the sine function returning the snowmelt coefficient has annual periodicity

        # First and last day-of-year for the "summer icemelt season" (approx. 1/4 year over summer) in the Northern (N) and Southern (S) hemispheres
        self.icemelt_start_N, self.icemelt_end_N = 165, 257 # first and last day-of-year for summer icemelt in the Northern hemisphere
        self.icemelt_start_S, self.icemelt_end_S = 347, 74 # first and last day-of-year for summer icemelt in the Southern hemisphere (+182 days wrt N)

        self.var.IceDayDegrees = 2 * self.var.SnowDayDegrees
        # days of summer (15th June-15th Sept.) to degrees: the sine function returning the icemelt coeff. has 1/2 year period, so it makes 1/2 cycle in 1 summer season
        self.var.SnowSeason = loadmap('SnowSeasonAdj') * 0.5
        # default value of range  of seasonal melt factor is set to 1
        # 0.5 x range of sinus function [-1,1]
        self.var.TempSnow = loadmap('TempSnow')
        self.var.SnowFactor = loadmap('SnowFactor')
        self.var.SnowMeltCoef = loadmap('SnowMeltCoef')
        self.var.TempMelt = loadmap('TempMelt')

        SnowCoverAInit = loadmap('SnowCoverAInitValue')
        SnowCoverBInit = loadmap('SnowCoverBInitValue')
        SnowCoverCInit = loadmap('SnowCoverCInitValue')
        self.var.SnowCoverS = [SnowCoverAInit, SnowCoverBInit, SnowCoverCInit]

        # initial snow depth in elevation zones A, B, and C, respectively  [mm]
        self.var.SnowCoverInit = (SnowCoverAInit + SnowCoverBInit + SnowCoverCInit) / 3
        # Pixel-average initial snow cover: average of values in 3 elevation zones
        self.var.SnowCover = globals.inZero.copy()


# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def dynamic(self):
        """ dynamic part of the snow module
        """
        self.var.Snow = globals.inZero.copy()
        self.var.Rain = globals.inZero.copy()
        self.var.SnowMelt = globals.inZero.copy()
        self.var.SnowCover = globals.inZero.copy()

        # Snowmelt
        hemisphere_N = self.var.lat_rad > 0
        snowmelt_coeff = np.sin(np.radians((self.var.CalendarDay - 81) * self.var.SnowDayDegrees))

        SeasSnowMeltCoef = self.var.SnowSeason * np.where(hemisphere_N, snowmelt_coeff, -snowmelt_coeff) + self.var.SnowMeltCoef # N and S hemispheres have opposite-sign cycles

        # Icemelt
	    #####################################################
        # # Bugfix from Emiliano - uncomment to make it work
        # # Check if the current day is in the "summer icemelt season" for the Northern (N) and Southern (S) hemispheres
        is_summer_icemelt_N = (self.var.CalendarDay > self.icemelt_start_N) & (self.var.CalendarDay < self.icemelt_end_N)
        is_summer_icemelt_S = (self.var.CalendarDay > self.icemelt_start_S) | (self.var.CalendarDay < self.icemelt_end_S)
        # Icemelt coefficient: the sine function is the same for both hemispheres due to the imposed 1/2 periodicity; the mask is shifted 6 months
        _ice_melt_coeff = np.sin(np.radians((self.var.CalendarDay - self.icemelt_start_N) * self.var.IceDayDegrees))
        ice_melt_coeff_N = _ice_melt_coeff if is_summer_icemelt_N else 0
        ice_melt_coeff_S = _ice_melt_coeff if is_summer_icemelt_S else 0
        SummerSeason = np.where(hemisphere_N, ice_melt_coeff_N, ice_melt_coeff_S)

        #####################################################
        # Bugfix for global simulations - LA 17/7/2018 - uncomment to make it work
        # This bugfix is on hold waiting for approval from main developer
        # Constants
        #SVE= 127.5     # Shifted Vernal Equinox (in days) = Vernal Equinox (20 March) + day shift to have the peak of Summer season on a desired day (9th August instead of 21 June)
        #Sc = 1361      # Solar constant (not really important in this case, as long as it is divided by a meaningful scale parameter)
        #SCp= 400       # Scale parameter of the SummerSeason coefficient
        #SHp= 0.6       # Shape parameter of the SummerSeason coefficient

        #LatRad = 2*np.pi*self.var.lat_rad/360  #Latitude in radians
        #Dec = np.arcsin(0.3987*np.sin(2*np.pi*(self.var.CalendarDay-SVE)/365.25)) #Declination (i.e., CalendarDay in radians)
        
        #H1 = np.arccos(np.sign(-1*np.tan(Dec)*np.tan(LatRad)))
        #H0 = np.arccos(-1*np.tan(Dec)*np.tan(LatRad))
        
        #SSc1=np.maximum(0,Sc/np.pi*(H1*np.sin(LatRad)*np.sin(Dec)+np.sin(H1)*np.cos(LatRad)*np.cos(Dec))/SCp-SHp)
        #SSc0=np.maximum(0,Sc/np.pi*(H0*np.sin(LatRad)*np.sin(Dec)+np.sin(H0)*np.cos(LatRad)*np.cos(Dec))/SCp-SHp)
        
        #mask = np.absolute(-1*np.tan(Dec)*np.tan(LatRad))>1
        # SScoef=np.where(mask,SSc1,SSc0) # org
        #SummerSeason=np.where(mask,SSc1,SSc0) # changed name back to SummerSeason

        ########################################################
        #Previous version with hardcoded dates:
        #if (self.var.CalendarDay > 165) and (self.var.CalendarDay < 260):
        #    SummerSeason = np.sin(math.radians((self.var.CalendarDay - 165) * self.var.IceDayDegrees))
        #else:
        #    SummerSeason = 0.0
        ########################################################


        for i in xrange(3):
            TavgS = self.var.Tavg + self.var.DeltaTSnow * (i - 1)
            # Temperature at center of each zone (temperature at zone B equals Tavg)
            # i=0 -> highest zone
            # i=2 -> lower zone
            SnowS = np.where(TavgS < self.var.TempSnow, self.var.SnowFactor * self.var.Precipitation, globals.inZero)
            # Precipitation is assumed to be snow if daily average temperature is below TempSnow
            # Snow is multiplied by correction factor to account for undercatch of
            # snow precipitation (which is common)
            RainS = np.where(TavgS >= self.var.TempSnow, self.var.Precipitation, globals.inZero)
            # if it's snowing then no rain
            SnowMeltS = (TavgS - self.var.TempMelt) * SeasSnowMeltCoef * (1 + 0.01 * RainS) * self.var.DtDay


            if i < 2:
                IceMeltS = self.var.Tavg * 7.0 * self.var.DtDay * SummerSeason
                # if i = 0 and 1 -> higher and middle zone
            else:
                IceMeltS = TavgS * 7.0 * self.var.DtDay * SummerSeason

            SnowMeltS = np.maximum(np.minimum(SnowMeltS + IceMeltS, self.var.SnowCoverS[i]), globals.inZero)
            self.var.SnowCoverS[i] = self.var.SnowCoverS[i] + SnowS - SnowMeltS

            self.var.Snow += SnowS
            self.var.Rain += RainS
            self.var.SnowMelt += SnowMeltS
            self.var.SnowCover += self.var.SnowCoverS[i]

        self.var.Snow /= 3
        self.var.Rain /= 3
        self.var.SnowMelt /= 3
        self.var.SnowCover /= 3

        self.var.TotalPrecipitation += self.var.Snow + self.var.Rain
        # total precipitation in pixel [mm]






"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *

# CM: new-style class in Python 2.x
class stateVar(object):

    """
    # ************************************************************
    # ***** ROUTING      *****************************************
    # ************************************************************
    """

    def __init__(self, stateVar_variable):
        self.var = stateVar_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
    def dynamic(self):
        """ reporting of state variables in dynamic part
        """
        try:
            EnKFset = option['EnKF']
        except:
            EnKFset = 0
        if EnKFset and self.var.currentTimeStep() in FilterSteps:
            sample = str(self.var.currentSampleNumber())
            dumpObject("StartDate", self.var.CalendarDayStart, sample)
            ## Snow
            dumpObject("SnowCover", self.var.SnowCoverS, sample)
            ## Soil Moisture
            dumpObject("Intercept", self.var.Interception, sample)
            dumpObject("CumIntercept", self.var.CumInterception, sample)
            dumpObject("Frost", self.var.FrostIndex, sample)
            dumpObject("W1a", self.var.W1a, sample)
            dumpObject("W1b", self.var.W1b, sample)
            dumpObject("W1", self.var.W1, sample)
            dumpObject("W2", self.var.W2, sample)
            dumpObject("DSLR", self.var.DSLR, sample)
            ## Groundwater
            dumpObject("UZ", self.var.UZ, sample)
            dumpObject("LZ", self.var.LZ, sample)
            ## Lakes
            if option['simulateLakes']:
                dumpObject("LakeStorageM3", self.var.LakeStorageM3CC, sample)
                dumpObject("LakeOutflow", self.var.LakeOutflow, sample)
            ## Routing
            dumpObject("ChanM3Kin", self.var.ChanM3Kin, sample)
            dumpObject("ChanQ", self.var.ChanQ, sample)
            dumpObject("ToChan", self.var.ToChanM3RunoffDt, sample)
            try:
                dumpObject("Chan2M3Kin", self.var.Chan2M3Kin, sample)
                dumpObject("Sideflow1Chan", self.var.Sideflow1Chan, sample)
            except:
                foo = 0
            ## Overland
            dumpObject("OFM3Direct", self.var.OFM3Direct, sample)
            dumpObject("OFM3Other", self.var.OFM3Other, sample)
            dumpObject("OFM3Forest", self.var.OFM3Forest, sample)
            ### Reservoirs
            if option['simulateReservoirs']:
                dumpObject("ReservoirStorageM3CC", self.var.ReservoirStorageM3CC, sample)
                dumpObject("ReservoirFill", self.var.ReservoirFill, sample)
                dumpObject("QResOutM3Dt", self.var.QResOutM3Dt, sample)
            try:
                dumpObject("Tss", self.var.Tss, sample)
            except:
                foo = 0
	    dumpObject("cdfFlag", globals.cdfFlag, sample)

    def resume(self):

        sample = str(self.var.currentSampleNumber())

        updateVec = self.var.getStateVector(sample)

        self.var.CalendarDayStart = loadObject("StartDate", sample)
        ## Snow
        self.var.SnowCoverS = loadObject("SnowCover", sample)
        ## Soil Moisture
        self.var.Interception = loadObject("Intercept", sample)
        self.var.CumInterception = loadObject("CumIntercept", sample)
        self.var.FrostIndex = loadObject("Frost", sample)
        self.var.W1a = loadObject("W1a", sample)
        self.var.W1b = loadObject("W1b", sample)
        self.var.W1 = loadObject("W1", sample)
        self.var.W2 = loadObject("W2", sample)
        self.var.DSLR = loadObject("DSLR", sample)
        ## Groundwater
        self.var.UZ = loadObject("UZ", sample)
        self.var.LZ = loadObject("LZ", sample)
        ## Lakes
        if option['simulateLakes']:
            self.var.LakeStorageM3CC = loadObject("LakeStorageM3", sample)
            self.var.LakeOutflow = loadObject("LakeOutflow", sample)
        ## Routing
        self.var.ChanM3Kin = loadObject("ChanM3Kin", sample)
        self.var.ChanQ = loadObject("ChanQ", sample)
        self.var.ToChanM3RunoffDt = loadObject("ToChan", sample)
        try:
            self.var.Chan2M3Kin = loadObject("Chan2M3Kin", sample)
            self.var.Sideflow1Chan = loadObject("Sideflow1Chan", sample)
        except:
            foo = 0
        ## Overland
        self.var.OFM3Direct = loadObject("OFM3Direct", sample)
        self.var.OFM3Other = loadObject("OFM3Other", sample)
        self.var.OFM3Forest = loadObject("OFM3Forest", sample)
        ### Reservoirs
        if option['simulateReservoirs']:
            self.var.ReservoirStorageM3CC = loadObject("ReservoirStorageM3CC", sample)
            self.var.ReservoirFill = loadObject("ReservoirFill", sample)
            self.var.QResOutM3Dt = loadObject("QResOutM3Dt", sample)
        try:
            self.output_module.var.Tss = loadObject("Tss", sample)
        except:
            foo = 0
        globals.cdfFlag = loadObject("cdfFlag", sample)



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

Parallel implementation of the kinematic wave routing.
This module provides tools to pre-process the flow direction matrix, and the kinematicWave class that
solves the kinematic wave equation. Most functions and methods call functions in the Cython module
kinematic_wave_parallel_tools.pyx, for performance reasons.
Parallelisation is achieved by grouping the pixels in ordered sets. Within each set, provided that
pixelsm in previous sets have already been routed, pixels can be routed independently of each other
and thus in parallel. For further details, see the method kinematicWave._setRoutingOrders.
"""

__author__ = "Emiliano Gelati"
__contact__ = "emiliano.gelati@ec.europa.eu"
__version__ = "1.0"
__date__ = "2016/06/01"

import os
import numpy as np
import pandas as pd
import numexpr as nx
from multiprocessing import cpu_count
from platform import system

# IMPORT THE PARALLELISE KINEMATIC WAVE RUTING MODULE: IF IT WAS NOT COMPILED ON THE CURRENT MACHINE, ROUTING IS EXECUTED SERIALLY
# If the binary .so file does not exist or is not newer than the source .pyx, then the Cython module is imported directly from the source.
# For safety against binary files compiled on other machines and copied here, the binary must be at least 10 seconds younger than the source (see line 13).
# Importing directly from the source prevents using OpenMP multithreading. In such case, the routing is executed serially.

WINDOWS_OS = system() == "Windows"
ROOT = os.path.join(os.path.dirname(os.path.realpath(__file__)), "kinematic_wave_parallel_tools")
SRC = ROOT + ".pyx"
BIN = ROOT + (".pyd" if WINDOWS_OS else ".so")


if not os.path.exists(BIN) or (os.path.exists(BIN) and os.path.exists(SRC) and os.stat(BIN).st_mtime < os.stat(SRC).st_mtime):
    import pyximport                         # Activate the direct import from source of Cython modules.
    setup_args = {"script_args": ["--compiler=mingw32"]} if WINDOWS_OS else None # Extra compiler argument under Windows
    pyximport.install(setup_args=setup_args) # If this is executed, the binary .so file will be ignored and the routing is executed serially.
    print("""WARNING:\nThe Cython module {} has not been compiled on the current machine (to compile, see instructions in the module's docstring).
The kinematic wave routing is executed serially (not in parallel).""".format(SRC))

from lisflood.hydrological_modules import kinematic_wave_parallel_tools as kwpt


# -------------------------------------------------------------------------------------------------
# CONSTANTS
# -------------------------------------------------------------------------------------------------

IX_ADDS = np.array([(1, 0), (1, 1), (0, 1), (-1, 1), (-1, 0), (-1, -1), (0, -1), (1, -1)]) # flow directions (row and column shifts in coordinate mesh)
SEA_CODE = 0 # Value for sea pixel in LISFLOOD flow direction matrix
FLOW_CODE = [2, 3, 6, 9, 8, 7, 4, 1, 5] # Flow directions according to LISFLOOD encoding (see IX_ADDS for directions expressed as row/column shifts)



# -------------------------------------------------------------------------------------------------
# FUNCTIONS
# -------------------------------------------------------------------------------------------------

def rebuildFlowMatrix(compressed_encoded_ldd, land_mask):
    """"""
    encoded_flow_dir = SEA_CODE * np.ones(land_mask.shape, compressed_encoded_ldd.dtype)
    encoded_flow_dir[land_mask] = compressed_encoded_ldd
    return encoded_flow_dir

def decodeFlowMatrix(encoded):
    '''Converts LISFLOOD flow matrix values (FLOW_CODE) to flow directions (0 to 7, see IX_ADDS)'''
    decoded = np.empty(encoded.shape, int)
    encoded[encoded == SEA_CODE] = FLOW_CODE[8]
    for old, new in zip(FLOW_CODE, range(9)):
        decoded[encoded == old] = new
    return np.ascontiguousarray(decoded)

def streamLookups(flow_dir, land_mask):
    '''
    Compute the downstream lookup vector for a D8 water flow channel network,
    i.e. the adjecency list of the directed graph describing flow direction from each pixel.
    Arguments:
        flow_dir (numpy.ndarray): LISFLOOD flow matrix values (FLOW_CODE).
        land_mask (numpy.ndarray): land mask on coordinate mesh.
    Returns:
        downstream lookup (numpy.ndarray): for each pixel, index of downstream pixel (-1 = none); size = num_pixels
        upstream lookup (numpy.ndarray): each row gives the immediately upstream pixels (-1 = fill value); size = num_pixels, max_ups_pixs <= 8
    '''
    flow_dir[~land_mask] = 8 # exceeds number of rows of IX_ADDS
    num_pixs = land_mask.sum()
    land_points = -np.ones(land_mask.shape, int)
    land_points[land_mask] = np.arange(num_pixs, dtype=int)
    downstream_lookup, upstream_lookup = kwpt.upDownLookups(flow_dir, np.ascontiguousarray(land_mask).astype(np.uint8), land_points, num_pixs, IX_ADDS)
    max_num_ups_pixs = max(1, np.any(upstream_lookup != -1, 0).sum()) # maximum number of upstreams pixels
    return downstream_lookup, np.ascontiguousarray(upstream_lookup[:,:max_num_ups_pixs]).astype(int) # astype for cython import in windows (see below)

def topoDistFromSea(downstream_lookup, upstream_lookup):
    """"""
    num_pixels = downstream_lookup.size
    topological_distance = - np.ones(num_pixels, int)
    is_outlet = np.zeros(num_pixels, bool)
    is_outlet[downstream_lookup == -1] = True
    distance = 1
    while (topological_distance == -1).any():
        to_track = np.logical_and(is_outlet, topological_distance == -1)
        topological_distance[to_track] = distance
        next_ups = np.unique(upstream_lookup[to_track])
        next_ups = next_ups[next_ups != -1]
        is_outlet[next_ups] = True
        distance += 1
    return topological_distance



# -------------------------------------------------------------------------------------------------
# CLASS
# -------------------------------------------------------------------------------------------------

class kinematicWave:
    """"""

    def __init__(self, compressed_encoded_ldd, land_mask, alpha_channel, beta, space_delta, time_delta, num_threads, alpha_floodplains=None):
        """"""
        # Parameters for the solution of the discretised Kinematic wave continuity equation
        self.space_delta = space_delta
        self.beta = beta
        self.inv_beta = 1 / beta
        self.b_minus_1 = beta - 1
        self.a_dx_div_dt_channel = alpha_channel * space_delta / time_delta
        self.b_a_dx_div_dt_channel = beta * self.a_dx_div_dt_channel
        # Set number of parallel threads (openMP)
        self.num_threads = int(num_threads) if 0 < num_threads <= cpu_count() else cpu_count()
        # If split-routing (floodplains)
        if alpha_floodplains is not None:
            self.a_dx_div_dt_floodplains = alpha_floodplains * space_delta / time_delta
            self.b_a_dx_div_dt_floodplains = beta * self.a_dx_div_dt_floodplains
        # Process flow direction matrix: downstream and upstream lookups, and routing orders
        flow_dir = decodeFlowMatrix(rebuildFlowMatrix(compressed_encoded_ldd, land_mask))
        self.downstream_lookup, self.upstream_lookup = streamLookups(flow_dir, land_mask)
        self.num_upstream_pixels = (self.upstream_lookup != -1).sum(1).astype(int) # astype for cython import in windows (to avoid 'long long' buffer dtype mismatch)
        # Routing order: decompose domain into batches; within each batch, pixels can be routed in parallel
        self._setRoutingOrders()

    def _setRoutingOrders(self):
        """Compute the kinematic wave routing order. Pixels are grouped in sets with the same order.
        Pixels in the same se are independent and can be routed in parallel. Sets must be processed in series, starting from order 0.
        Pixels are ordered topologically starting from the outlets, as in:
        Liu et al. (2014), A layered approach to parallel computing for spatially distributed hydrological modeling,
        Environmental Modelling & Software 51, 221-227.
        Order MAX is given to pixels with no downstream relations (outlets); order MAX-1 is given to
        pixels whose downstream pixels are all of order MAX; and so on."""
        ocean_topo_distance = topoDistFromSea(self.downstream_lookup, self.upstream_lookup)
        routing_order = ocean_topo_distance.max() - ocean_topo_distance
        self.pixels_ordered = pd.DataFrame({"pixels": np.arange(routing_order.size), "order": routing_order})
        try:
            self.pixels_ordered = self.pixels_ordered.sort_values(["order", "pixels"]).set_index("order").squeeze()
        except: # FOR COMPATIBILITY WITH OLDER PANDAS VERSIONS
            self.pixels_ordered = self.pixels_ordered.sort(["order", "pixels"]).set_index("order").squeeze()
        order_counts = self.pixels_ordered.groupby(self.pixels_ordered.index).count()
        stop = order_counts.cumsum()
        self.order_start_stop = np.column_stack((np.append(0, stop[:-1]), stop)).astype(int) # astype for cython import in windows (see above)
        self.pixels_ordered = self.pixels_ordered.values.astype(int) # astype for cython import in windows (see above)

    def kinematicWaveRouting(self, discharge, specific_lateral_inflow, section="main_channel"):
        """Kinematic wave routing: wrapper around kinematic_wave_parallel_tools.kinematicWave"""
        # Lateral inflow (m3 s-1)
        lateral_inflow = nx.evaluate("q * dx", local_dict={"q": specific_lateral_inflow, "dx": self.space_delta})
        # Choose between main channel and floodplain routing
        if section == "main_channel":
            a_dx_div_dt = self.a_dx_div_dt_channel
            b_a_dx_div_dt = self.b_a_dx_div_dt_channel
        elif section == "floodplains":
            a_dx_div_dt = self.a_dx_div_dt_floodplains
            b_a_dx_div_dt = self.b_a_dx_div_dt_floodplains
        else:
            raise Exception("The section parameter must be either 'main_channel' or 'floodplain'!")
        # Constant term in f(x) evaluation for Newton-Raphson method
        local_dict = {"a_dx_div_dt": a_dx_div_dt, "Qold": discharge, "b": self.beta, "lateral_inflow": lateral_inflow}
        constant = nx.evaluate("a_dx_div_dt * Qold ** b + lateral_inflow", local_dict=local_dict)
        # Solve the Kinematic wave equation
        kwpt.kinematicRouting(discharge, lateral_inflow, constant, self.upstream_lookup,\
                              self.num_upstream_pixels, self.pixels_ordered, self.order_start_stop,\
                              self.beta, self.inv_beta, self.b_minus_1, a_dx_div_dt, b_a_dx_div_dt, self.num_threads)


"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

 ######################################################################

 ##       ####  ######  ######## ##        #######   #######  ########
 ##        ##  ##    ## ##       ##       ##     ## ##     ## ##     ##
 ##        ##  ##       ##       ##       ##     ## ##     ## ##     ##
 ##        ##   ######  ######   ##       ##     ## ##     ## ##     ##
 ##        ##        ## ##       ##       ##     ## ##     ## ##     ##
 ##        ##  ##    ## ##       ##       ##     ## ##     ## ##     ##
 ######## ####  ######  ##       ########  #######   #######  ########

######################################################################
"""

# to work with the new grid engine JRC - workaround with error on pyexpat
from pyexpat import *
import xml.dom.minidom
from netCDF4 import Dataset
from pcraster import *
from pcraster.framework import *

# ---------------------------

import os
import sys

current_dir = os.path.dirname(os.path.abspath(__file__))
src_dir = os.path.join(current_dir, '../')
if os.path.exists(src_dir):
    sys.path.append(src_dir)

from Lisflood_initial import *
from Lisflood_dynamic import *
from Lisflood_monteCarlo import *
from Lisflood_EnKF import *
from metainfo import __authors__, __version__, __date__, __status__


class LisfloodModel(LisfloodModel_ini, LisfloodModel_dyn, LisfloodModel_monteCarlo, LisfloodModel_EnKF):
    """ Joining the initial and the dynamic part
        of the Lisflood model
    """

# ==================================================
# ============== LISFLOOD execute ==================
# ==================================================


def Lisfloodexe(settings, optionxml):

    # read options and bindings and launch Lisflood model computation
    # returns option binding and ReportSteps - global dictionaries
    optionBinding(settings, optionxml)
    # read all the possible option for modelling and for generating output
    # read the settingsfile with all information about the catchments(s)
    # and the choosen option for mdelling and output
    bindkey = sorted(binding.keys())

    #os.chdir(outputDir[0])
    # this prevent from using relative path in settings!


    checkifDate('StepStart','StepEnd')
    # check 'StepStart' and 'StepEnd' to be >0 and 'StepStart'>'StepEnd'
    # return modelSteps

    # remove steps from ReportSteps that are not included in simulation period
    for key in list(ReportSteps.keys()):  ## creates a list of all keys
        ReportSteps[key] = [x for x in ReportSteps[key] if x >= modelSteps[0]]
        ReportSteps[key] = [x for x in ReportSteps[key] if x <= modelSteps[1]]

    if option['InitLisflood']: print "INITIALISATION RUN"

    #print start step and end step
    print "Start Step - End Step: ",modelSteps[0]," - ", modelSteps[1]
    print "Start Date - End Date: ",inttoDate(modelSteps[0]-1,Calendar(binding['CalendarDayStart']))," - ",\
        inttoDate(modelSteps[1]-1,Calendar(binding['CalendarDayStart']))


    if Flags['loud']:
        # print state file date
        print "State file Date: ",
        try:
            print inttoDate(Calendar(binding["timestepInit"]), Calendar(binding['CalendarDayStart']))
        except:
            print Calendar(binding["timestepInit"])

        # CM: print start step and end step for reporting model state maps
        print "Start Rep Step  - End Rep Step: ", ReportSteps['rep'][0], " - ", ReportSteps['rep'][-1]
        print "Start Rep Date  - End Rep Date: ", inttoDate(Calendar(ReportSteps['rep'][0] - 1),
                                                            Calendar(binding['CalendarDayStart'])), \
            " - ", inttoDate(Calendar(ReportSteps['rep'][-1] - 1), Calendar(binding['CalendarDayStart']))

        # messages at model start
        print"%-6s %10s %11s\n" %("Step","Date","Discharge"),


    # Lisflood is an instance of the class LisfloodModel
    # LisfloodModel includes 2 methods : initial and dynamic (formulas applied at every timestep)
    Lisflood = LisfloodModel()
    # stLisflood is an instance of the class DynamicFramework

    stLisflood = DynamicFramework(Lisflood, firstTimestep=modelSteps[0], lastTimeStep=modelSteps[1])
    stLisflood.rquiet = True
    stLisflood.rtrace = False


    """
    ----------------------------------------------
    Monte Carlo and Ensemble Kalman Filter setting
    ----------------------------------------------
    """
    # Ensemble Kalman filter
    try:
        EnKFset = option['EnKF']
    except:
        EnKFset = 0
    # MonteCarlo
    try:
        MCset = option['MonteCarlo']
    except:
        MCset = 0
    if option['InitLisflood']:
        MCset = 0
        EnKFset = 0
    if EnKFset and not MCset:
        msg = "Trying to run EnKF with only 1 ensemble member \n"
        raise LisfloodError(msg)
    if EnKFset and FilterSteps[0] == 0:
        msg = "Trying to run EnKF without filter timestep specified \nRunning LISFLOOD in Monte Carlo mode \n"
        print LisfloodWarning(msg)
        EnKFset = 0
    if MCset and EnsMembers[0] <= 1:
        msg = "Trying to run Monte Carlo simulation with only 1 member \nRunning LISFLOOD in deterministic mode \n"
        print LisfloodWarning(msg)
        MCset = 0
    if MCset:
        mcLisflood = MonteCarloFramework(stLisflood, nrSamples=EnsMembers[0])
        if nrCores[0] > 1:
            mcLisflood.setForkSamples(True, nrCPUs=nrCores[0])
        if EnKFset:
            kfLisflood = EnsKalmanFilterFramework(mcLisflood)
            kfLisflood.setFilterTimesteps(FilterSteps)
            print LisfloodRunInfo(mode = "Ensemble Kalman Filter", outputDir = outputDir[0], Steps = len(FilterSteps), ensMembers=EnsMembers[0], Cores=nrCores[0])
            kfLisflood.run()
        else:
            print LisfloodRunInfo(mode = "Monte Carlo", outputDir = outputDir[0], ensMembers=EnsMembers[0], Cores=nrCores[0])
            mcLisflood.run()
    else:
        """
        ----------------------------------------------
        Deterministic run
        ----------------------------------------------
        """
        print LisfloodRunInfo(mode = "Deterministic", outputDir = outputDir[0])
    # run of the model inside the DynamicFramework
        stLisflood.run()
    # cProfile.run('stLisflood.run()')
    # python -m cProfile -o  l1.pstats lisf1.py settingsNew3.xml
    # gprof2dot -f pstats l1.pstats | dot -Tpng -o callgraph.png

    if Flags['printtime']:
        print "\n\nTime profiling"
        print "%2s %-17s %10s %8s" %("No","Name","time[s]","%")
        div = 1
        timeSum = np.array(timeMesSum)
        if MCset:
            div = div * EnsMembers[0]
        if EnKFset:
            div = div * (len(FilterSteps)+1)
        if EnKFset or MCset:
            timePrint = np.zeros(len(timeSum)/div)
            for i in range(len(timePrint)):
                timePrint[i] = np.sum(timeSum[range(i, len(timeSum), len(timeSum)/div)])
        else:
            timePrint = timeSum
        for i in xrange(len(timePrint)):
            print "%2i %-17s %10.2f %8.1f"  %(i,timeMesString[i],timePrint[i],100 * timePrint[i] / timePrint[-1])

# ==================================================
# ============== USAGE ==============================
# ==================================================


def usage():
    """ prints some lines describing how to use this program
        which arguments and parameters it accepts, etc
    """

    print 'LisfloodPy - Lisflood using pcraster Python framework'
    print 'Authors: ', __authors__
    print 'Version: ', __version__
    print 'Date: ', __date__
    print 'Status: ', __status__
    print """
    Arguments list:
    settings.xml     settings file

    -q --quiet       output progression given as .
    -v --veryquiet   no output progression is given
    -l --loud        output progression given as time step, date and discharge
    -c --check       input maps and stack maps are checked, output for each input map BUT no model run
    -h --noheader    .tss file have no header and start immediately with the time series
    -t --printtime   the computation time for hydrological modules are printed
    -d --debug       debug outputs
    """
    sys.exit(1)


def headerinfo():

   print "LisfloodPy ", __version__, " ", __date__,
   print """
Water balance and flood simulation model for large catchments\n
(C) Institute for Environment and Sustainability
    Joint Research Centre of the European Commission
    TP122, I-21020 Ispra (Va), Italy\n"""


# ==================================================
# ============== MAIN ==============================
# ==================================================


def main():
    # if arguments are missing display usage info

    if len(sys.argv) < 2:
        usage()

    # global settings
    # global optionxml
    optionxml = get_optionxml_path()

    # setting.xml file
    settings = sys.argv[1]

    # arguments list
    args = sys.argv[2:]

    # Flags - set model behavior (quiet,veryquiet, loud, checkfiles, noheader,printtime,debug)
    globalFlags(args)
    # setting of global flag e.g checking input maps, producing more output information
    if not (Flags['veryquiet']) and not (Flags['quiet']):
        headerinfo()
    Lisfloodexe(settings, optionxml)


def get_optionxml_path():
    # read OptionTserieMaps.xml in the same folder as Lisflood main (lisf1.py)
    LF_Path = os.path.dirname(os.path.abspath(sys.argv[0]))
    # OptionTserieMaps.xml file
    optionxml = os.path.normpath(os.path.join(LF_Path, "OptionTserieMaps.xml"))
    if not os.path.exists(optionxml):
        optionxml = os.path.normpath(os.path.join(current_dir, "OptionTserieMaps.xml"))
    return optionxml


if __name__ == "__main__":
    sys.exit(main())



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *


class inflow(object):

    """
     # ************************************************************
     # ***** READ INFLOW HYDROGRAPHS (OPTIONAL)****************
     # ************************************************************
     # If option "inflow" is set to 1 the inflow hydrograph code is used
     # otherwise dummy code is used
    """

    def __init__(self, inflow_variable):
        self.var = inflow_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
    def initial(self):
        """ initial part of the inflow module
        """
        # ************************************************************
        # ***** INFLOW INIT
        # ************************************************************

        if option['inflow']:
            self.var.InflowPoints = loadmap('InflowPoints')
            self.var.QInM3Old = np.where(self.var.InflowPoints>0,self.var.ChanQ * self.var.DtSec,0)

            # read inflow map
            inflowmapprc = loadmap('InflowPoints',pcr=True)
            inflowmapnp = (inflowmapprc, -9999)
            inflowmapnp = np.where(inflowmapnp>0,inflowmapnp,0)

            # get outlets ids from outlets map
            inflowId = np.unique(inflowmapnp)
            # drop negative values (= missing data in pcraster map)
            inflowId = inflowId[inflowId > 0]

            # read tss ids from tss file
            tssId = read_tss_header(binding['QInTS'])

            # create a dictionary of tss id : tss id index
            id_dict = {}
            for i in range(len(tssId)):
                id_dict[tssId[i]] = tssId.index(tssId[i]) +1

            # remove inflow point if not available in tss file
            for i in range(len(inflowId)):
                if inflowId[i] in tssId:
                    pass
                else:
                    id_dict[inflowId[i]] = 0
                    msg = "Inflow point was removed ID:", str(inflowId[i]) ,"\n"
                    print LisfloodWarning(msg)


            # substitute indexes to id in map
            self.var.InflowPointsMap = np.copy(inflowmapnp)
            for k, v in id_dict.iteritems(): self.var.InflowPointsMap[inflowmapnp==k] = v

            # convert map to pcraster format
            # self.var.InflowPointsMap = decompress(self.var.InflowPointsMap)
            self.var.InflowPointsMap = numpy2pcr(Nominal, self.var.InflowPointsMap, -9999)
            tempnpinit = pcr2numpy(self.var.InflowPointsMap,-9999)
            pass

            #self.var.QInM3Old = cover(ifthen(defined(self.var.InflowPoints), self.var.ChanQ * self.var.DtSec), scalar(0.0))
            # Initialising cumulative output variables
            # These are all needed to compute the cumulative mass balance error

#        self.var.QInDt = globals.inZero.copy()
        # inflow substep amount

    def dynamic_init(self):
        """ dynamic part of the inflow module
            init inflow before sub step routing
        """

        # ************************************************************
        # ***** INLETS INIT
        # ************************************************************
        if option['inflow']:
            self.var.QDelta = (self.var.QInM3 - self.var.QInM3Old) * self.var.InvNoRoutSteps
            # difference between old and new inlet flow  per sub step
            # in order to calculate the amount of inlet flow in the routing loop

    def dynamic(self):
        """ dynamic part of the inflow module
        """

        if option['inflow']:
            QIn = timeinputscalar(binding['QInTS'], self.var.InflowPointsMap)

            # Get inflow hydrograph at each inflow point [m3/s]
            QIn = compressArray(QIn)
            QIn[np.isnan(QIn)]=0
            self.var.QInM3 = QIn * self.var.DtSec
            #self.var.QInM3 = cover(QIn * self.var.DtSec, 0)
            # Convert to [m3] per time step
            self.var.TotalQInM3 += self.var.QInM3
            # Map of total inflow from inflow hydrographs [m3]


    def dynamic_inloop(self,NoRoutingExecuted):
        """ dynamic part of the inflow routine
           inside the sub time step routing routine
        """

        # ************************************************************
        # ***** INLFLOW **********************************************
        # ************************************************************
        if option['inflow']:
            self.var.QInDt = (self.var.QInM3Old + (NoRoutingExecuted + 1) * self.var.QDelta) * self.var.InvNoRoutSteps
            # flow from inlets per sub step



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""


from lisflood.global_modules.add1 import *


class soil(object):

    """
    # ************************************************************
    # ***** SOIL         *****************************************
    # ************************************************************
    """

    def __init__(self, soil_variable):
        self.var = soil_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the soil module
        """
        def splitlanduse(array1, array2=None, array3=None):
            """ splits maps into the 3 different land use types - other , forest, irrigation
            """
            if array2 is None:
                array2 = array1
            if array3 is None:
                array3 = array1
            return [array1, array2, array3]

        # ------------------------------------------------
        # Fraction of Landuse
        # using maps instead of table to be more sensitive to landuse change
        # Soil properties for percentage of pixel area without forest, water ,
        # impervious soil

        #np.seterr(invalid='ignore',divide='ignore')

#        self.var.RiceFraction = loadmap('RiceFraction')
#        self.var.ForestFraction = loadmap('ForestFraction')
#        self.var.OtherFraction = loadmap('OtherFraction')
#        self.var.IrrigationFraction = loadmap('IrrigationFraction')
#        self.var.DirectRunoffFraction = loadmap('DirectRunoffFraction')
#        self.var.WaterFraction = loadmap('WaterFraction')

        self.var.OtherFraction += self.var.RiceFraction
           # for the moment rice is treated as other fraction	        if option['varfractionwater']:
            # if the fraction of water varies then the other fraction are stored
        self.var.WaterFractionBase = self.var.WaterFraction.copy()
        self.var.OtherFractionBase = self.var.OtherFraction.copy()
        self.var.IrrigationFractionBase = self.var.IrrigationFraction.copy()
        self.var.ForestFractionBase = self.var.ForestFraction.copy()
        self.var.DirectRunoffFractionBase = self.var.DirectRunoffFraction.copy()

        self.var.SoilFraction = self.var.ForestFraction +  self.var.OtherFraction + self.var.IrrigationFraction
        self.var.PermeableFraction = 1 - self.var.DirectRunoffFraction - self.var.WaterFraction
        # Permeable fraction of pixel

        # Soil Depth 1st layer [mm]
        # soil depth top layer for every landuse but forest, impervious soil
        # and water

        self.var.SoilDepth1a = defsoil('SoilDepth1','SoilDepth1Forest')
        self.var.SoilDepth1b = defsoil('SoilDepth2','SoilDepth2Forest')
        self.var.SoilDepth2  = defsoil('SoilDepth3','SoilDepth3Forest')

# ----------------- miscParameters ---------------------------

        self.var.CourantCrit = loadmap('CourantCrit')
        self.var.LeafDrainageK = np.minimum(self.var.DtDay * (1 / loadmap('LeafDrainageTimeConstant')), 1)
        # Reservoir constant for interception store
        # Minimize statement ONLY needed if LeafDrainageTimeConstant
        # becomes greater than DtDay (unlikely, but possible for time steps
        # greater than daily)

        self.var.AvWaterThreshold = loadmap('AvWaterRateThreshold') * self.var.DtDay
        # Converts AvWaterRateThreshold from  [mm/day]
        # to [mm] per time step


# ************************************************************
# ***** LAND USE RELATED MAPS ********************************
# ************************************************************

        # using maps instead of table to be more sensitive to landuse change
        # Soil properties for percentage of pixel area without forest, water ,
        # impervious soil
        #self.var.CropCoef = [loadmap('MapCropCoef'), loadmap('MapForestCropCoef')]
        self.var.CropCoef = defsoil('MapCropCoef','MapForestCropCoef','MapIrrigationCropCoef')

        # crop coefficients for each land use class

        self.var.CropGroupNumber = defsoil('MapCropGroupNumber','MapForestCropGroupNumber','MapIrrigationCropGroupNumber')

        # crop group numbers for soil water depletion factor
        # self.var.NManning = [loadmap('MapN'), loadmap('MapForestN'), 0.02]
        self.var.NManning = defsoil('MapN','MapForestN', 0.02)
        # Manning's roughness coefficient for each land use class
        # third manning is from direct runoff
        # self.var.NManningDirect=scalar(0.02)


# ************************************************************
# ***** MAPS DERIVED FROM SOIL TEXTURE AND SOIL DEPTH ********
# ************************************************************

        # HYPRES Parameters for TopSoil:

        self.var.KSat1a = defsoil('MapKSat1', 'MapKSat1Forest')
        self.var.KSat1b = defsoil('MapKSat2', 'MapKSat2Forest')
        self.var.KSat2 = defsoil('MapKSat3')
        # Saturated conductivity for other land use and forest
        Lambda1a = defsoil('MapLambda1','MapLambda1Forest')
        Lambda1b = defsoil('MapLambda2','MapLambda2Forest')
        Lambda2 = defsoil('MapLambda3')
        # Pore-size index (for Van Genuchten parameters)
        GenuAlpha1a = defsoil('MapGenuAlpha1','MapGenuAlpha1Forest')
        GenuAlpha1b = defsoil('MapGenuAlpha2','MapGenuAlpha2Forest')
        GenuAlpha2 = defsoil('MapGenuAlpha3')
        # Van Genuchten Alpha coefficient
        ThetaS1a = defsoil('MapThetaSat1','MapThetaSat1Forest')
        ThetaS1b = defsoil('MapThetaSat2','MapThetaSat2Forest')
        ThetaS2 = defsoil('MapThetaSat3')
        # Soil moisture content at saturation [V/V]
        ThetaRes1a = defsoil('MapThetaRes1', 'MapThetaRes1Forest')
        ThetaRes1b = defsoil('MapThetaRes2', 'MapThetaRes2Forest')
        ThetaRes2 = defsoil('MapThetaRes3')
        # Soil moisture content at saturation [V/V]

        # GenuN1=Lambda+1
        GenuN1a = [x + 1 for x in Lambda1a]
        GenuN1b = [x + 1 for x in Lambda1b]
        GenuN2 = [x + 1 for x in Lambda2]
        # self.var.GenuM1=Lambda1/GenuN1
        self.var.GenuM1a = [x / y for x, y in zip(Lambda1a, GenuN1a)]
        self.var.GenuM1b = [x / y for x, y in zip(Lambda1b, GenuN1b)]
        self.var.GenuM2 = [x / y for x, y in zip(Lambda2, GenuN2)]
        # self.var.GenuInvM1=1/self.var.GenuM1
        self.var.GenuInvM1a = [1 / x for x in self.var.GenuM1a]
        self.var.GenuInvM1b = [1 / x for x in self.var.GenuM1b]
        self.var.GenuInvM2 = [1 / x for x in self.var.GenuM2]
        # self.var.GenuInvN1=1/GenuN1
        self.var.GenuInvN1a = [1 / x for x in GenuN1a]
        self.var.GenuInvN1b = [1 / x for x in GenuN1b]
        self.var.GenuInvN2 = [1 / x for x in GenuN2]
        # GenuInvAlpha1=1/GenuAlpha1
        self.var.GenuInvAlpha1a = [1 / x for x in GenuAlpha1a]
        self.var.GenuInvAlpha1b = [1 / x for x in GenuAlpha1b]
        self.var.GenuInvAlpha2 = [1 / x for x in GenuAlpha2]
        # Van Genuchten n and m coefficients
        # In previous versions both n and m were read from separate tables.
        # Since both are calculated from same pore-size index (Lambda=n-1),
        # only table of Lambda is used here (Lambda is also more commonly
        # reported in the literature, which makes any comparison easier)
        # Inverse values computed once here for improved performance
        # (inverse of N and Alpha only needed if option simulatePF is used)

        # self.var.WS1=ThetaS1*self.var.SoilDepth1
        self.var.WS1a = [x * y for x, y in zip(ThetaS1a, self.var.SoilDepth1a)]
        self.var.WS1b = [x * y for x, y in zip(ThetaS1b, self.var.SoilDepth1b)]
        self.var.WS2 = [x * y for x, y in zip(ThetaS2, self.var.SoilDepth2)]
        self.var.WS1 = np.add(self.var.WS1a, self.var.WS1b)

        # self.var.WRes1=ThetaRes1*self.var.SoilDepth1
        self.var.WRes1a = [x * y for x, y in zip(ThetaRes1a, self.var.SoilDepth1a)]
        self.var.WRes1b = [x * y for x, y in zip(ThetaRes1b, self.var.SoilDepth1b)]
        self.var.WRes1 = np.add(self.var.WRes1a, self.var.WRes1b)
        self.var.WRes2 = [x * y for x, y in zip(ThetaRes2, self.var.SoilDepth2)]

        # self.var.WS1WRes=self.var.WS1-self.var.WRes1
        self.var.WS1WResa = [x - y for x, y in zip(self.var.WS1a, self.var.WRes1a)]
        self.var.WS1WResb = [x - y for x, y in zip(self.var.WS1b, self.var.WRes1b)]
        self.var.WS2WRes = [x - y for x, y in zip(self.var.WS2, self.var.WRes2)]

        # Express volume fractions in [mm] water slice

        # self.var.WFC1=self.var.WRes1+(self.var.WS1-self.var.WRes1)/((1+(GenuAlpha1*100)**GenuN1)**self.var.GenuM1)
        self.var.WFC1a = map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 100)
                       ** gn) ** gm), self.var.WRes1a, self.var.WS1a, GenuAlpha1a, GenuN1a, self.var.GenuM1a)
        self.var.WFC1b = map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 100)
                       ** gn) ** gm), self.var.WRes1b, self.var.WS1b, GenuAlpha1b, GenuN1b, self.var.GenuM1b)
        self.var.WFC1 = np.add(self.var.WFC1a, self.var.WFC1b)
        self.var.WFC2 = map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 100)
                       ** gn) ** gm), self.var.WRes2, self.var.WS2, GenuAlpha2, GenuN2, self.var.GenuM2)
        # Soil moisture at field capacity (pF2, 100 cm) [mm water slice]
        # Mualem equation (van Genuchten, 1980)

        self.var.WPF3a = map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 1000)
                        ** gn) ** gm), self.var.WRes1a, self.var.WS1a, GenuAlpha1a, GenuN1a, self.var.GenuM1a)
        self.var.WPF3b = map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 1000)
                        ** gn) ** gm), self.var.WRes1b, self.var.WS1b, GenuAlpha1b, GenuN1b, self.var.GenuM1b)
        self.var.WPF3 = np.add(self.var.WPF3a, self.var.WPF3b)

        # Soil moisture at field capacity (pF3, 1000 cm) [mm water slice]
        # Mualem equation (van Genuchten, 1980)        # self.var.WWP1=self.var.WRes1+(self.var.WS1-self.var.WRes1)/  ((1+(GenuAlpha1*(10**4.2))**GenuN1)**self.var.GenuM1)
        self.var.WWP1a = map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 15000)
                       ** gn) ** gm), self.var.WRes1a, self.var.WS1a, GenuAlpha1a, GenuN1a, self.var.GenuM1a)
        self.var.WWP1b = map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 15000)
                       ** gn) ** gm), self.var.WRes1b, self.var.WS1b, GenuAlpha1b, GenuN1b, self.var.GenuM1b)
        self.var.WWP1 = np.add(self.var.WWP1a, self.var.WWP1b)
        self.var.WWP2 = map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 15000)
                       ** gn) ** gm), self.var.WRes2, self.var.WS2, GenuAlpha2, GenuN2, self.var.GenuM2)
        # Soil moisture at wilting point (pF4.2, 10**4.2 cm) in [mm] water slice
        # Mualem equation (van Genuchten, 1980)

        self.var.PoreSpaceNotZero1a = splitlanduse(np.bool8((self.var.SoilDepth1a[0] != 0) & (self.var.WS1a[0] != 0)),
             np.bool8((self.var.SoilDepth1a[1] != 0) & (self.var.WS1a[1] != 0)),
             np.bool8((self.var.SoilDepth1a[2] != 0) & (self.var.WS1a[2] != 0)))
        self.var.PoreSpaceNotZero1b = splitlanduse(np.bool8((self.var.SoilDepth1b[0] != 0) & (self.var.WS1b[0] != 0)),
             np.bool8((self.var.SoilDepth1b[1] != 0) & (self.var.WS1b[1] != 0)),
             np.bool8((self.var.SoilDepth1b[2] != 0) & (self.var.WS1b[2] != 0)))
        self.var.PoreSpaceNotZero2 = splitlanduse(np.bool8((self.var.SoilDepth2[0] != 0) & (self.var.WS2[0] != 0)),
             np.bool8((self.var.SoilDepth2[1] != 0) & (self.var.WS2[1] != 0)),
             np.bool8((self.var.SoilDepth2[2] != 0) & (self.var.WS2[2] != 0)))

#        self.var.PoreSpaceNotZero1 = [np.bool8((self.var.SoilDepth1[0] != 0) & (self.var.WS1[0] != 0)),
#            np.bool8((self.var.SoilDepth1[1] != 0) & (self.var.WS1[1] != 0))]
#        self.var.PoreSpaceNotZero2 = [np.bool8((self.var.SoilDepth2[0] != 0) & (self.var.WS2[0] != 0)),
#            np.bool8((self.var.SoilDepth2[1] != 0) & (self.var.WS2[1] != 0))]
        # Flag that is Boolean 1 if pixel has (any) soil moisture storage space.
        # Needed to avoid divisions by zero in soil moisture calculations

        # ***** INITIAL VALUES for soil

        ThetaInit1aValue = defsoil('ThetaInit1Value','ThetaForestInit1Value','ThetaIrrigationInit1Value')
        ThetaInit1bValue = defsoil('ThetaInit2Value','ThetaForestInit2Value','ThetaIrrigationInit2Value')
        ThetaInit2Value = defsoil('ThetaInit3Value','ThetaForestInit3Value','ThetaIrrigationInit3Value')

        WInit1a = splitlanduse(globals.inZero.copy())
        WInit1b = splitlanduse(globals.inZero.copy())
        WInit2 =  splitlanduse(globals.inZero.copy())
        WInit1a[0] = np.where(ThetaInit1aValue[0] == -9999, self.var.WFC1a[0], ThetaInit1aValue[0] * self.var.SoilDepth1a[0])
        WInit1a[1] = np.where(ThetaInit1aValue[1] == -9999, self.var.WFC1a[1], ThetaInit1aValue[1] * self.var.SoilDepth1a[1])
        WInit1a[2] = np.where(ThetaInit1aValue[2] == -9999, self.var.WFC1a[2], ThetaInit1aValue[2] * self.var.SoilDepth1a[2])
        WInit1b[0] = np.where(ThetaInit1bValue[0] == -9999, self.var.WFC1b[0], ThetaInit1bValue[0] * self.var.SoilDepth1b[0])
        WInit1b[1] = np.where(ThetaInit1bValue[1] == -9999, self.var.WFC1b[1], ThetaInit1bValue[1] * self.var.SoilDepth1b[1])
        WInit1b[2] = np.where(ThetaInit1bValue[2] == -9999, self.var.WFC1b[2], ThetaInit1bValue[2] * self.var.SoilDepth1b[2])
        WInit2[0] = np.where(ThetaInit2Value[0] == -9999, self.var.WFC2[0], ThetaInit2Value[0] * self.var.SoilDepth2[0])
        WInit2[1] = np.where(ThetaInit2Value[1] == -9999, self.var.WFC2[1], ThetaInit2Value[1] * self.var.SoilDepth2[1])
        WInit2[2] = np.where(ThetaInit2Value[2] == -9999, self.var.WFC2[2], ThetaInit2Value[2] * self.var.SoilDepth2[2])
        # Initial soil moisture in permeable fraction of pixel
        # Set soil to field capacity if initial value
        # in settings file equals -9999
        # IMPORTANT: WInit1 and WInit2 represent the soil moisture in the
        # *permeable* fraction of the pixel *only* (since all soil moisture-related
        # calculations are done for permeable fraction only!).

        WInit1a[0] = np.where(self.var.PoreSpaceNotZero1a[0], WInit1a[0], globals.inZero)
        WInit1a[1] = np.where(self.var.PoreSpaceNotZero1a[1], WInit1a[1], globals.inZero)
        WInit1a[2] = np.where(self.var.PoreSpaceNotZero1a[2], WInit1a[2], globals.inZero)
        WInit1b[0] = np.where(self.var.PoreSpaceNotZero1b[0], WInit1b[0], globals.inZero)
        WInit1b[1] = np.where(self.var.PoreSpaceNotZero1b[1], WInit1b[1], globals.inZero)
        WInit1b[2] = np.where(self.var.PoreSpaceNotZero1b[2], WInit1b[2], globals.inZero)
        WInit2[0] = np.where(self.var.PoreSpaceNotZero2[0], WInit2[0], globals.inZero)
        WInit2[1] = np.where(self.var.PoreSpaceNotZero2[1], WInit2[1], globals.inZero)
        WInit2[2] = np.where(self.var.PoreSpaceNotZero2[2], WInit2[2], globals.inZero)
        # Set to zero if soil depth is zero
        self.var.W1a = WInit1a
        self.var.W1b = WInit1b
        self.var.W1 = np.add(self.var.W1a, self.var.W1b)
        self.var.W2 = WInit2
        # Set soil moisture to initial value

        self.var.Sat1a = splitlanduse(globals.inZero.copy())
        self.var.Sat1b = splitlanduse(globals.inZero.copy())
        self.var.Sat1 = splitlanduse(globals.inZero.copy())
        self.var.Sat2 = splitlanduse(globals.inZero.copy())


# ************************************************************
# ***** REPEATEDLY USED EXPRESSIONS IN XINANJIANG
# ***** INFILTRATION MODEL
# ************************************************************
        self.var.b_Xinanjiang = loadmap('b_Xinanjiang')
        self.var.PowerInfPot = (self.var.b_Xinanjiang + 1) / self.var.b_Xinanjiang
        # Power in infiltration equation
        # self.var.StoreMaxPervious=self.var.WS1/(self.var.b_Xinanjiang+1)
#        self.var.StoreMaxPervious = [x / (self.var.b_Xinanjiang + 1) for x in self.var.WS1a]

        self.var.StoreMaxPervious = [x / (self.var.b_Xinanjiang + 1) for x in self.var.WS1]
        # Maximum soil moisture storage in pervious fraction of
        # pixel (expressed as [mm] water slice)


# ********************************************
# ****  PowerPrefFlow
# ********************************************
        self.var.PowerPrefFlow = loadmap('PowerPrefFlow')

# ************************************************************
# ***** INITIAL VALUES
# ************************************************************
# Inputs in waterbalance model and/or initial assumption
        # CMmod
        # DSLRInit = defsoil('DSLRInitValue', 'DSLRForestInitValue','DSLRIrrigationInitValue')

        DSLRInit = defsoil('DSLRInitValue', 'DSLRForestInitValue', 'DSLRIrrigationInitValue')
        self.var.DSLR = splitlanduse(*[np.maximum(d_i, 1) for d_i in DSLRInit])
        # Days since last rainfall
        # Impose lower limit of 1 because otherwise
        # square root may be taken out of a negative number for Dtsec lt 86400!

        # CumInterceptionInit=ifthen(defined(MaskMap),scalar(loadmap('CumIntInitValue')))
        self.var.CumInterception = defsoil('CumIntInitValue', 'CumIntForestInitValue', 'CumIntIrrigationInitValue')
        #self.var.CumInterception = [CumInterceptionInit[0], CumInterceptionInit[1]]
        # initial cumulative interception, [mm]

        # Initialising cumulative output variables
        # These are all needed to compute the cumulative mass balance error


        self.var.TotalPrecipitation = globals.inZero.copy()
        # precipitation [mm]
        self.var.TaCUM = globals.inZero.copy()
        # Cumulative transpiration [mm]
        self.var.TaInterceptionCUM = globals.inZero.copy()
        # Cumulative evaporation from interception store [mm]
        self.var.ESActCUM = globals.inZero.copy()
        # Cumulative evaporation [mm]

        self.var.SoilMoistureStressDays = splitlanduse(globals.inZero.copy())
        # number of days in simulation with soil moisture stress (days)

        self.var.Theta1a = splitlanduse(globals.inZero.copy())
        self.var.Theta1b = splitlanduse(globals.inZero.copy())
        self.var.Theta2 =  splitlanduse(globals.inZero.copy())

        self.var.TaInterception = defsoil(globals.inZero.copy())
        self.var.Interception = defsoil(globals.inZero.copy())
        self.var.LeafDrainage = defsoil(globals.inZero.copy())

        self.var.Ta = splitlanduse(globals.inZero.copy())
        self.var.ESAct = splitlanduse(globals.inZero.copy())
        self.var.PrefFlow = splitlanduse(globals.inZero.copy())
        self.var.Infiltration = splitlanduse(globals.inZero.copy())
        self.var.SeepTopToSubA = splitlanduse(globals.inZero.copy())
        self.var.SeepTopToSubB = splitlanduse(globals.inZero.copy())
        self.var.SeepSubToGW = splitlanduse(globals.inZero.copy())

        self.var.Theta = splitlanduse(globals.inZero.copy())
        self.var.AvailableWaterForInfiltration = splitlanduse(globals.inZero.copy())
        self.var.RWS = splitlanduse(globals.inZero.copy())


# ************************************************************
# ***** INITIALIZATION OF IMPERVIOUS SOIL     ****************
# ************************************************************

        self.var.CumInterSealed = loadmap('CumIntSealedInitValue')
        # cumulative depression storage
        self.var.SMaxSealed = loadmap('SMaxSealed')
        # maximum depression storage for water on impervious surface
        # which is not immediately causing surface runoff  [mm]

        if option['drainedIrrigation']:
            self.var.DrainedFraction = loadmap('DrainedFraction')
            # drained fraction: water from irrigated water is directly send to channel
            # depending on the drained fraction
# ************************************************************
# ***** INITIALIZATION OF PF calculation      ****************
# ************************************************************

        if option['simulatePF']:
            self.var.HeadMax = loadmap('HeadMax')
            self.var.pF0 = defsoil(globals.inZero.copy())
            self.var.pF1 = defsoil(globals.inZero.copy())
            self.var.pF2 = defsoil(globals.inZero.copy())

    def dynamic_perpixel(self):
        """ dynamic part of the soil module
            Calculation per Pixel
        """
        self.var.TaInterceptionAll = self.var.deffraction(self.var.TaInterception) + self.var.DirectRunoffFraction * self.var.TASealed
        self.var.TaInterceptionCUM += self.var.TaInterceptionAll
        # Cumulative evaporation of intercepted water [mm]

        self.var.TaPixel = self.var.deffraction(self.var.Ta)
        # pixel-average transpiration in [mm] per timestep
        # (no transpiration from direct runoff fraction)
        self.var.TaCUM += self.var.TaPixel
        self.var.ESActPixel = self.var.deffraction(self.var.ESAct) + self.var.WaterFraction * self.var.EWaterAct
        # Pixel-average soil evaporation in [mm] per time step
        # (no evaporation from direct runoff fraction)
        self.var.ESActCUM += self.var.ESActPixel


        self.var.PrefFlowPixel = self.var.deffraction(self.var.PrefFlow)
        # Pixel-average preferential flow in [mm] per timestep
        # (no preferential flow from direct runoff fraction)
        self.var.InfiltrationPixel = self.var.deffraction(self.var.Infiltration)
        # Pixel-average infiltration in [mm] per timestep
        # (no infiltration in direct runoff fraction)


        self.var.Theta[0] = self.var.OtherFraction * (self.var.W1a[0] + self.var.W1b[0] + self.var.W2[0]) / (
            self.var.SoilDepth1a[0] + self.var.SoilDepth1b[0] + self.var.SoilDepth2[0])
        self.var.Theta[1] = self.var.ForestFraction * (self.var.W1a[1] + self.var.W1b[1] + self.var.W2[1]) / (
            self.var.SoilDepth1a[1] + self.var.SoilDepth1b[1] + self.var.SoilDepth2[1])
        self.var.Theta[2] = self.var.IrrigationFraction * (self.var.W1a[2] + self.var.W1b[2] + self.var.W2[2]) / (
            self.var.SoilDepth1a[2] + self.var.SoilDepth1b[2] + self.var.SoilDepth2[2])
        self.var.ThetaAll = np.where(self.var.SoilFraction > 0, (self.var.Theta[
            0] + self.var.Theta[1] + self.var.Theta[2]) / self.var.SoilFraction,0.0)

        self.var.SeepTopToSubPixelA = self.var.deffraction(self.var.SeepTopToSubA)
        self.var.SeepTopToSubPixelB = self.var.deffraction(self.var.SeepTopToSubB)
        self.var.SeepSubToGWPixel = self.var.deffraction(self.var.SeepSubToGW)
        # Pixel-average seepage values in [mm] per timestep
        # (no seepage from direct runoff fraction)



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from global_modules.output import *


class LisfloodModel_EnKF(DynamicModel, MonteCarloModel, EnKfModel):

    """ LISFLOOD initial part
        same as the PCRaster script -initial-
        this part is to initialize the variables
        it will call the initial part of the hydrological modules
    """

    def __init__(self):
        """ init part of the initial part
            defines the mask map and the outlet points
            initialization of the hydrological modules
        """
        DynamicModel.__init__(self)
        MonteCarloModel.__init__(self)
        EnKfModel.__init__(self)

    def setState(self):
        ### Create some random model simulations
        ### This is done for each ensemble member individually
        values = np.random.normal(0,1,5)
        #print str("Model") + str(values)
        ## Return values to EnKF framework
        return values

    def setObservations(self):
        ## Create some random observations
        values = np.random.normal(0,1,5)

        # creating the observation matrix (nrObservations x nrSamples)
        # here without added noise
        observations = numpy.array([values,]*self.nrSamples()).transpose()

        # creating the covariance matrix (nrObservations x nrObservations)
        # here just random values
        covariance = numpy.random.random((5, 5))

        ## Return observations to EnKF framework
        self.setObservedMatrices(observations, covariance)

    def resume(self):
        sample = str(self.currentSampleNumber())

        updateVec = self.getStateVector(sample)

        #print str("update") + str(updateVec)
        self.stateVar_module.resume()



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""


from lisflood.global_modules.add1 import *


class opensealed(object):

    """
    # ************************************************************
    # ***** SOIL LOOP    *****************************************
    # ************************************************************
    """

    def __init__(self, opensealed_variable):
        self.var = opensealed_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def dynamic(self):
        """ dynamic part of the open water and sealed soil module
        """
        # ************************************************************
        # ***** ACTUAL EVAPORATION FROM OPEN WATER AND SEALED SOIL ***
        # ************************************************************

        self.var.RainSnowmelt = np.maximum(self.var.Rain + self.var.SnowMelt, globals.inZero)
        # Water available for the impervious soil and water bodies during this
        # timestep [mm]      #


        self.var.EWaterAct = np.minimum(self.var.EWRef, self.var.RainSnowmelt)
        self.var.EWaterAct = np.maximum(self.var.EWaterAct * 1.0, globals.inZero)
        # actual evaporation from water is potential evaporation of water bodies


        self.var.InterSealed =np.maximum(self.var.SMaxSealed - self.var.CumInterSealed, globals.inZero)
        self.var.InterSealed = np.minimum(self.var.InterSealed, self.var.RainSnowmelt)
        # Interception (in [mm] per time step);
        # to simulate initial loss and depression storage
        self.var.CumInterSealed += self.var.InterSealed


        self.var.TASealed = np.maximum(np.minimum(self.var.CumInterSealed, self.var.EWRef), globals.inZero)
        # evaporation of initial loss and depression storage using potential
        # evaporation of water bodies
        self.var.CumInterSealed = np.maximum(self.var.CumInterSealed - self.var.TASealed, globals.inZero)
        # evaporated water is subtracted from Cumulative depression storage;


        self.var.DirectRunoff = self.var.DirectRunoffFraction * (self.var.RainSnowmelt - self.var.InterSealed) + self.var.WaterFraction *  (self.var.RainSnowmelt - self.var.EWaterAct)
        # Direct runoff during this time step [mm] (added to surface runoff later)
        # but overland routing is done separately for forest, water and sealed
        # and remaing areas




"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""


from lisflood.global_modules.add1 import *


class indicatorcalc(object):

    """
    # ************************************************************
    # ***** Indicator calculation ************************************
    # ************************************************************
    """

    def __init__(self, indicatorcalc_variable):
        self.var = indicatorcalc_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the indicator calculation module
        """

        self.var.monthend = False
        self.var.yearend = False
        try:
            self.var.DefineEndofYear = int(binding['DefineEndofYear'])
        except:
            self.var.DefineEndofYear = 365

        if option['indicator']:
            # population per pixel
            self.var.Population = loadmap('Population')
            # map to mask out deserts and high mountains (to cover ETdif map, otherwise Sahara etc would pop out; meant as a drought indicator
            self.var.LandUseMask = loadmap('LandUseMask')

            if option['wateruse']:
                self.var.RegionPopulation = np.take(np.bincount(self.var.WUseRegionC, weights=self.var.Population), self.var.WUseRegionC)
                # population sum in Regions

                # set to 0 at start
                self.var.DayCounter = 0
                self.var.MonthETpotMM = globals.inZero.copy()
                self.var.MonthETactMM = globals.inZero.copy()

                self.var.MonthWDemandM3 = globals.inZero.copy()
                self.var.MonthWAbstractionM3 = globals.inZero.copy()
                self.var.MonthWConsumptionM3 = globals.inZero.copy()
                self.var.MonthDisM3 = globals.inZero.copy()
                self.var.MonthInternalFlowM3 = globals.inZero.copy()
                self.var.MonthExternalInflowM3 = globals.inZero.copy()
                self.var.RegionMonthIrrigationShortageM3 = globals.inZero.copy()
                self.var.MonthWaterAbstractedfromLakesReservoirsM3 = globals.inZero.copy()

                self.var.LakeStorageM3 = globals.inZero.copy()

    def dynamic(self):
        """ dynamic part of the indicator calculation module
        """

        if option['TransientLandUseChange']:
            self.var.Population = readnetcdf(binding['PopulationMaps'], self.var.currentTimeStep())
            self.var.RegionPopulation = np.take(np.bincount(self.var.WUseRegionC, weights=self.var.Population), self.var.WUseRegionC)
            # population sum in Regions

        if option['wateruse'] and option['indicator']:
            # check if it is the last monthly or annual time step
            next_date_time = self.var.CalendarDate + datetime.timedelta(seconds=int(binding["DtSec"]))
            self.var.monthend = next_date_time.month != self.var.CalendarDate.month
            self.var.yearend = next_date_time.year != self.var.CalendarDate.year
            # sum up every day
            self.var.DayCounter   = self.var.DayCounter + 1.0
            self.var.MonthETpotMM   = self.var.MonthETpotMM + self.var.ETRef
            self.var.MonthETactMM   = self.var.MonthETactMM + self.var.deffraction(self.var.TaInterception) + self.var.TaPixel + self.var.ESActPixel

            if option['openwaterevapo']:
                self.var.MonthETactMM += self.var.EvaAddM3 * self.var.M3toMM

            self.var.MonthETdifMM = np.maximum((self.var.MonthETpotMM - self.var.MonthETactMM)*self.var.LandUseMask,globals.inZero)
            # ; land use mask can be used to mask out deserts and high mountains, where no agriculture is possible

            self.var.MonthWDemandM3 = self.var.MonthWDemandM3 + self.var.TotalDemandM3
            self.var.MonthWAbstractionM3 = self.var.MonthWAbstractionM3 + self.var.TotalAbstractionFromSurfaceWaterM3 + self.var.ReservoirAbstractionM3 + self.var.LakeAbstractionM3 + self.var.TotalAbstractionFromGroundwaterM3
            self.var.MonthWConsumptionM3 = self.var.MonthWConsumptionM3  + self.var.WUseAddM3 + self.var.ReservoirAbstractionM3 + self.var.LakeAbstractionM3 + self.var.TotalAbstractionFromGroundwaterM3
            self.var.MonthDisM3 = self.var.MonthDisM3 + self.var.ChanQAvg * self.var.DtSec

            self.var.MonthWaterAbstractedfromLakesReservoirsM3 = self.var.MonthWaterAbstractedfromLakesReservoirsM3 + self.var.ReservoirAbstractionM3 + self.var.LakeAbstractionM3

            self.var.RegionMonthIrrigationShortageM3 = self.var.RegionMonthIrrigationShortageM3 + self.var.AreatotalIrrigationShortageM3

            # INTERNAL FLOW
            self.var.MonthInternalFlowM3 = self.var.MonthInternalFlowM3 + self.var.ToChanM3Runoff

            # --------------------------------------------------------------------------

            if self.var.monthend:

                # INTERNAL FLOW
                if option['simulateReservoirs'] or option['simulateLakes']:
                    # available LakeStorageM3 and ReservoirStorageM3 for potential abstraction at end of month in region
                    self.var.RegionMonthReservoirAndLakeStorageM3 = np.take(np.bincount(self.var.WUseRegionC, weights=(self.var.ReservoirStorageM3 + self.var.LakeStorageM3)), self.var.WUseRegionC)

                # monthly abstraction from lakes and reservoirs
                self.var.RegionMonthWaterAbstractedfromLakesReservoirsM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthWaterAbstractedfromLakesReservoirsM3),self.var.WUseRegionC)

                #PerMonthInternalFlowM3  = areatotal(cover(decompress(self.var.MonthInternalFlow),0.0),wreg)
                self.var.RegionMonthInternalFlowM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthInternalFlowM3),self.var.WUseRegionC)
                # note Reservoir and Lake storage need to be taken into account seperately

                # EXTERNAL FLOW
                wreg = decompress(self.var.WUseRegionC)
                     # to pcraster map because of the following expression!
                self.var.RegionMonthExternalInflowM3 = compressArray(areatotal(cover(ifthen(self.var.WaterRegionInflowPoints != 0, upstream(self.var.LddStructuresKinematic,decompress(self.var.MonthDisM3))),0),wreg))

                self.var.RegionMonthWAbstractionM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthWAbstractionM3),self.var.WUseRegionC)
                self.var.RegionMonthWConsumptionM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthWConsumptionM3),self.var.WUseRegionC)
                self.var.RegionMonthWDemandM3      = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthWDemandM3),self.var.WUseRegionC)

                # Calculation of WEI: everything in m3, totalled over the water region

                UpstreamInflowM3 = self.var.RegionMonthExternalInflowM3
                LocalFreshwaterM3 = self.var.RegionMonthInternalFlowM3
                # still to be decided if reservoir and lake water availability is added here
                LocalTotalWaterDemandM3 = self.var.RegionMonthWDemandM3
                RemainingDemandM3 = np.maximum(LocalTotalWaterDemandM3 - LocalFreshwaterM3,0.0)
                # this is the demand that cannot be met by local water supply
                UpstreamInflowUsedM3 = np.minimum(RemainingDemandM3,UpstreamInflowM3)
                # the amount of upstream water an area really depends on
                FossilGroundwaterUsedM3 = np.maximum(RemainingDemandM3 - UpstreamInflowUsedM3,0.0)
                # the amount of water that cannot be met locally nor with international water
                # likely this is the amount of water drawn from deep groundwater

                self.var.WEI_Cns = np.where((UpstreamInflowM3+LocalFreshwaterM3) > 0.0,self.var.RegionMonthWConsumptionM3 / (UpstreamInflowM3+LocalFreshwaterM3),0.0)
                self.var.WEI_Abs = np.where((UpstreamInflowM3+LocalFreshwaterM3) > 0.0,self.var.RegionMonthWAbstractionM3 / (UpstreamInflowM3+LocalFreshwaterM3),0.0)
                self.var.WEI_Dem = np.where((UpstreamInflowM3+LocalFreshwaterM3) > 0.0,self.var.RegionMonthWDemandM3 / (UpstreamInflowM3+LocalFreshwaterM3),0.0)

                self.var.WaterSustainabilityIndex =  np.where(LocalTotalWaterDemandM3 > 0.0, FossilGroundwaterUsedM3 / (LocalTotalWaterDemandM3+1),0.0)
                # De Roo 2015: WTI, if above zero, indicates that situation is unsustainable
                # if index is 0 means sustainable situtation: no groundwater or desalination water used
                # if index is 1 means area relies completely on groundwater or desalination water
                # the '+1' is to prevent division by small values, leading to very large and misleading indicator values

                self.var.WaterDependencyIndex =  np.where(LocalTotalWaterDemandM3 > 0.0,  UpstreamInflowUsedM3 / (LocalTotalWaterDemandM3+1),0.0)
                # De Roo 2015: WDI, dependency on upstreamwater, as a fraction of the total local demand
                # the '+1' is to prevent division by small values, leading to very large and misleading indicator values

                self.var.WaterSecurityIndex = np.where(UpstreamInflowM3 > 0.0,  UpstreamInflowUsedM3 / (UpstreamInflowM3+1),0.0)
                # De Roo 2015: WSI, indicates the vulnerability to the available upstream water;
                # if only 10% of upstream inflow is use, WSI would be 0.1 indicating low vulnerability
                # if WSI is close to 1, situation is very vulnerable
                # the '+1' is to prevent division by small values, leading to very large and misleading indicator values

                self.var.FalkenmarkM3Capita1 =  np.where(self.var.RegionPopulation > 0.0,self.var.RegionMonthInternalFlowM3*12/self.var.RegionPopulation,globals.inZero.copy())
                self.var.FalkenmarkM3Capita2 =  np.where(self.var.RegionPopulation > 0.0,LocalFreshwaterM3*12/self.var.RegionPopulation,globals.inZero.copy())
                self.var.FalkenmarkM3Capita3 =  np.where(self.var.RegionPopulation > 0.0,(UpstreamInflowM3+LocalFreshwaterM3)*12/self.var.RegionPopulation,globals.inZero.copy())
                # FalkenmarkM3Capita1 = (TotalRegionMonthToChanM3)/TotalTimeSteps*365.25/(PopulationRegion+0.0001);
                # FalkenmarkM3Capita2 = (TotalRegionMonthToChanM3+TotalRegionMonthReservoirLakeAbstractionM3)/TotalTimeSteps*365.25/(PopulationRegion+0.0001);
                # FalkenmarkM3Capita3 = (TotalRegionMonthToChanM3+TotalRegionMonthReservoirLakeAbstractionM3+TotalRegionMonthExternalInflowM3)/TotalTimeSteps*365.25/(PopulationRegion+0.0001);
                # report(decompress(self.var.WEI_Use),'E:/test.map')

            # --------------------------------------------------------------------------

            if self.var.yearend:
                x=1




# --------------------------------------------------------------------------
    def dynamic_setzero(self):
        """ dynamic part of the indicator calculation module
            which set the monthly (yearly) values back to start
        """

        if option['wateruse'] and option['indicator']:

            if self.var.monthend:
            # set to 0 at the end of a month
                self.var.DayCounter = 0
                self.var.MonthETpotMM =   globals.inZero.copy()
                self.var.MonthETactMM =   globals.inZero.copy()
                self.var.MonthWDemandM3 = globals.inZero.copy()
                self.var.MonthWAbstractionM3 = globals.inZero.copy()
                self.var.MonthWConsumptionM3 = globals.inZero.copy()
                self.var.MonthDisM3 =     globals.inZero.copy()
                self.var.MonthInternalFlowM3 = globals.inZero.copy()
                self.var.MonthExternalInflowM3 = globals.inZero.copy()
                self.var.RegionMonthIrrigationShortageM3 = globals.inZero.copy()
                self.var.MonthWaterAbstractedfromLakesReservoirsM3 = globals.inZero.copy()

            if self.var.yearend:
                x=1



"""
Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

---------------------------------------------------------------------------------------------------------------------------------------
Use python setup.y upload to publish versioned tags and pypi package

Manually step by step:

1. python setup.py sdist

2a. To upload new package on PyPi Test:
twine upload --repository-url https://test.pypi.org/legacy/ dist/*

2b. To upload new package on PyPi:
twine upload dist/*

3a. Test package install
pip install --index-url https://test.pypi.org/simple/ lisflood-model==2.8.14

3b. In prod:
pip install lisflood-model
"""

import os
import sys
from glob import glob
from shutil import rmtree

from setuptools import find_packages, Extension, setup, Command, __path__ as setuppath

pip_package = os.path.join(setuppath[0], '../lisflood')
if os.path.exists(pip_package) and pip_package in sys.path:
    sys.path.remove(pip_package)

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(current_dir, './src/'))

from lisflood.metainfo import __version__, __authors__

try:
    # noinspection PyUnresolvedReferences
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

extension_ext = 'pyx' if USE_CYTHON else 'c'

ext_modules = [
    Extension(
        "lisflood.hydrological_modules.kinematic_wave_parallel_tools",
        ["src/lisflood/hydrological_modules/kinematic_wave_parallel_tools.{}".format(extension_ext)],
        extra_compile_args=["-O3", "-ffast-math", "-march=native", "-fopenmp"],
        extra_link_args=["-fopenmp"]
    )
]

if USE_CYTHON:
    ext_modules = cythonize(ext_modules)

readme_file = os.path.join(current_dir, 'README.md')
with open(readme_file, 'r') as f:
    long_description = f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Publish lisflood-model package.'
    user_options = []

    @staticmethod
    def print_console(s):
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.print_console('Removing previous builds...')
            rmtree(os.path.join(current_dir, 'dist'))
        except OSError:
            pass

        self.print_console('Building Source and Wheel (universal) distribution...')
        os.system('{0} setup.py sdist'.format(sys.executable))

        self.print_console('Uploading the package to PyPI via Twine...')
        os.system('twine upload dist/*')

        self.print_console('Pushing git tags...')
        os.system('git tag v{0}'.format(__version__))
        os.system('git push --tags')

        sys.exit()


setup(
    name='lisflood-model',
    version=__version__,
    package_dir={'lisflood': 'src/lisflood/'},
    py_modules=[os.path.splitext(os.path.basename(path))[0] for path in glob('src/*.py*')],
    include_package_data=True,
    package_data={'lisflood': ['*.xml']},
    packages=find_packages('src'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='LISFLOOD model python module',
    author=__authors__,
    author_email='domenico.nappo@ext.ec.europa.eu',
    keywords=['lisflood', 'lisvap', 'efas', 'glofas', 'ecmwf'],
    license='EUPL 1.2',
    url='https://github.com/ec-jrc/lisflood-model',
    download_url='https://github.com/ec-jrc/lisflood-code/archive/{}.tar.gz'.format(__version__),
    setup_requires=[
            # Setuptools 18.0 properly handles Cython extensions.
            'setuptools>=41.0',
            'numpy>=1.15',
            'cython',
    ],
    install_requires=[
        'python-dateutil', 'pytest', 'docopt',
        'numpy>=1.15', 'cython', 'netCDF4',
        'numexpr', 'pandas', 'xarray', 'pyproj', 'cftime',
    ],
    scripts=['bin/lisflood'],
    ext_modules=ext_modules,
    zip_safe=True,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    # setup.py publish to pypi.
    cmdclass={
        'upload': UploadCommand,
    },
)



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from pcraster import*
from pcraster.framework import *
import sys
import os
import string
import math


from lisflood.global_modules.zusatz import *
from lisflood.global_modules.add1 import *
from lisflood.global_modules.globals import *


class riceirrigation(object):

    """
    # ************************************************************
    # ***** Rice irrigation   ************************************
    # ************************************************************
    """

    def __init__(self, riceirrigation_variable):
        self.var = riceirrigation_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the rice irrigation module
        """
        self.var.PaddyRiceWaterAbstractionFromSurfaceWaterM3 = globals.inZero.copy()

        if option['riceIrrigation']:

        # ************************************************************
        # ***** PADDY RICE IRRIGATION AND ABSTRACTION ******************
        # ************************************************************

        # Additional water for paddy rice cultivation is calculated seperately, as well as additional open water evaporation from rice fields
            # self.var.RiceFlooding = loadmap('RiceFlooding') #original
            self.var.RiceFlooding = loadmap('RiceFlooding')
            # 10 mm for 10 days (total 10cm water)
            # self.var.RicePercolation = loadmap('RicePercolation') #original
            self.var.RicePercolation = loadmap('RicePercolation')
            # FAO: percolation for heavy clay soils: PERC = 2 mm/day

            self.var.RicePlantingDay1 = loadmap('RicePlantingDay1')
            # starting day (of the year) of first rice planting
            self.var.RiceHarvestDay1 = loadmap('RiceHarvestDay1')
            # starting day (of the year) of first rice harvest

            self.var.RicePlantingDay2 = loadmap('RicePlantingDay2')
             # starting day (of the year) of second rice planting
            self.var.RiceHarvestDay2 = loadmap('RiceHarvestDay2')
             # starting day (of the year) of 2nd rice harvest



    def dynamic(self):
        """ dynamic part of the rice irrigation routine
           inside the water abstraction routine
        """


        if option['riceIrrigation']:

        # water needed for paddy rice is assumed to consist of:
            # phase 1: field preparation: soil saturation (assumed to happen in 10 days, 20 days before planting)
            # phase 2: flood fields (assumed to happen in 10 days, 10 days before planting)
            # phase 3: planting, while keep constant water level during growing season (open water evaporation)
            # phase 4: stop keeping constant water level 20 days before harvest date
            # phase 5: start draining 10 days before harvest date

	        #RiceSoilSaturationDemandM3=(WS1-W1Loop1+WS2-W2Loop1)*RiceFraction*MMtoM3;
            # RiceSoilSaturationDemandM3 = (self.var.WS1[0]-self.var.W1[0] +  self.var.WS2[0]-self.var.W2[0]) * self.var.RiceFraction * self.var.MMtoM3 #original
            RiceSoilSaturationDemandM3 = (self.var.WS1[0] - self.var.W1[0] + self.var.WS2[0] - self.var.W2[0]) * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay
            # this part is using the whole other fraction to calculate the demand -> an rice only soil part is needed
            # RiceIrrigationDemandM3 unit is m3 per time interval [m3/dt]

            pl_20 = self.var.RicePlantingDay1-20
            pl_20 = np.where(pl_20 <0, 365+pl_20,pl_20)
            pl_10 = self.var.RicePlantingDay1-10
            pl_10 = np.where(pl_10 <0, 365+pl_10,pl_10)

            ha_20 = self.var.RiceHarvestDay1 - 20
            ha_20 = np.where(ha_20 < 0, 365+ha_20,ha_20)
            ha_10 = self.var.RiceHarvestDay1 - 10
            ha_10 = np.where(ha_10 < 0, 365+ha_10,ha_10)

             # for Europe ok, but for Global planting can be on the 330 and harvest on the 90, so harvest < planting
             # or riceplanting = 5 => riceplanting -20 =350 ==> riceplanting < riceplanting -20

            """ phase 1: field preparation: soil saturation (assumed to happen in 10 days, 20 days before planting)"""
            # RiceSoilSaturationM3=if((CalendarDay ge (RicePlantingDay1-20)) and (CalendarDay le (RicePlantingDay1-10)),0.1*RiceSoilSaturationDemandM3,0)
            RiceSoilSaturationM3 = np.where((self.var.CalendarDay >= pl_20) & (self.var.CalendarDay < pl_10), 0.1*RiceSoilSaturationDemandM3, globals.inZero)
            #RiceFloodingM3=if((CalendarDay ge (RicePlantingDay1-10)) and (CalendarDay le (RicePlantingDay1)),(RiceFlooding+EWRef)*RiceFraction*MMtoM3,0)

            RiceEva = self.var.EWRef - (self.var.ESAct[0]+self.var.Ta[0])
            RiceEva = np.maximum(RiceEva,0)
            RiceEvaporationDemandM3 = RiceEva * self.var.RiceFraction*self.var.MMtoM3   # m3 per time interval
               # should not happen, but just to be sure that this doesnt go <0
               # part of the evaporation is already taken out in soil module!
               # substracting the soil evaporation and transpiration which was already taken off in the soil module

            RiceFloodingDemandM3 = self.var.RiceFlooding * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay     # m3 per time interval

            """ phase 2: flood fields (assumed to happen in 10 days, 10 days before planting)"""
            
            # RiceFloodingM3 = np.where((self.var.CalendarDay >= pl_10) & (self.var.CalendarDay < self.var.RicePlantingDay1), (self.var.RiceFlooding+RiceEva)*self.var.RiceFraction*self.var.MMtoM3, globals.inZero) #original
            RiceFloodingM3 = np.where((self.var.CalendarDay >= pl_10) & (self.var.CalendarDay < self.var.RicePlantingDay1),RiceFloodingDemandM3+RiceEvaporationDemandM3,globals.inZero) #m3 per time interval
            # part of the evaporation is already taken out in soil module!
               # assumption is that a fixed water layer is kept on the rice fields, totalling RiceFlooding*10 in mmm (typically 50 or 100 mm)
               # application is spread out over 10 days
               # open water evaporation at the same time


            """ phase 3: planting, while keep constant water level during growing season (open water evaporation) """
            #RiceEvaporationM3=if((CalendarDay ge RicePlantingDay1) and (CalendarDay le (RiceHarvestDay1-20)),EWRef*RiceFraction*MMtoM3,0)
            # RiceEvaporationM3 = np.where((self.var.CalendarDay >= self.var.RicePlantingDay1) & (self.var.CalendarDay < ha_20), RiceEva * self.var.RiceFraction*self.var.MMtoM3 , globals.inZero) #original
            RiceEvaporationM3 = np.where((self.var.CalendarDay >= self.var.RicePlantingDay1) & (self.var.CalendarDay < ha_20),RiceEvaporationDemandM3, globals.inZero)  #m3 per time interval

            # substracting the soil evaporation which was already taken off in the soil module (also transpitation should be tyaken off )

            #RicePercolationM3=if((CalendarDay ge RicePlantingDay1) and (CalendarDay le (RiceHarvestDay1-20)),RicePercolation*RiceFraction*MMtoM3,0)
            RicePercolationDemandM3 = self.var.RicePercolation * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay   # m3 per time interval
            # RicePercolationM3 = np.where((self.var.CalendarDay >= self.var.RicePlantingDay1) & (self.var.CalendarDay < ha_20), self.var.RicePercolation*self.var.RiceFraction*self.var.MMtoM3, globals.inZero) #original
            RicePercolationM3 = np.where((self.var.CalendarDay >= self.var.RicePlantingDay1) & (self.var.CalendarDay < ha_20),RicePercolationDemandM3, globals.inZero)  # m3 per time interval
            # FAO: percolation for heavy clay soils: PERC = 2 mm/day
            self.var.PaddyRiceWaterAbstractionFromSurfaceWaterM3 = RiceSoilSaturationM3 + RiceFloodingM3 + RiceEvaporationM3 + RicePercolationM3    # m3 per time interval
               # m3 water needed for paddyrice

            #self.var.TotalAbstractionFromSurfaceWaterM3 = self.var.PaddyRiceWaterAbstractionFromSurfaceWaterM3
                
            """# phase 4: stop keeping constant water level 20 days before harvest date
                 phase 5: start draining 10 days before harvest date"""
            #RiceDrainageM3=if((CalendarDay ge (RiceHarvestDay1-10)) and (CalendarDay le RiceHarvestDay1),(WS1-WFC1+WS2-WFC2)*RiceFraction*MMtoM3,0)
            
            RiceDrainageDemandM3 = (self.var.WS1[0]-self.var.WFC1[0] + self.var.WS2[0]-self.var.WFC2[0]) * self.var.RiceFraction*self.var.MMtoM3 * self.var.DtDay   # m3 per time interval
            RiceDrainageM3 = np.where((self.var.CalendarDay >= ha_10) & (self.var.CalendarDay < self.var.RiceHarvestDay1),0.1 * RiceDrainageDemandM3,globals.inZero)
            # RiceDrainageM3 = np.where((self.var.CalendarDay >= ha_10) & (self.var.CalendarDay < self.var.RiceHarvestDay1),
            #                          0.1 * (self.var.WS1[0]-self.var.WFC1[0] + self.var.WS2[1]-self.var.WFC2[1]) * self.var.RiceFraction*self.var.MMtoM3,globals.inZero) #original

                # drainage until FC to soil/groundwater at end of season
                # assumption that the last weeks before harvest the 50mm water layer is completely evaporating
                # needs to be transported to channel system or being drained

            #UZLoop1 += cover((RiceDrainageM3+RicePercolationM3)*M3toMM/OtherFraction,0)
            self.var.UZ[0] += np.where(self.var.OtherFraction > 0.0,(RiceDrainageM3+RicePercolationM3)*self.var.M3toMM/self.var.OtherFraction,0.0)



 #           self.var.UZ[0] += (RicePercolationM3)*self.var.M3toMM/self.var.OtherFraction
              # drained water is added to Upper Zone








"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""


from pcraster import*
from pcraster.framework import *

from lisflood.global_modules.add1 import *
from lisflood.global_modules.globals import *


class transmission(object):

    """
    # ************************************************************
    # ***** Transmission loss ************************************
    # ************************************************************
    """

    def __init__(self, transmission_variable):
        self.var = transmission_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the transmission loss module
        """

        if option['TransLoss']:
            TransArea = loadmap('TransArea')
            self.var.TransSub = loadmap('TransSub')
            # downstream area taking into account for transmission loss
            self.var.UpAreaTrans = loadmap('UpAreaTrans')
            # upstream area
            self.var.UpTrans = np.where(self.var.UpAreaTrans >= TransArea,np.bool8(1),np.bool8(0))
            # Downstream taking into accound for transmission loss
            # if upstream area (the total one) is bigger than a threshold us
            # transmission loss
            self.var.TransPower1 = loadmap('TransPower1')
            self.var.TransPower2 = 1.0 / self.var.TransPower1
            # transmission loss function

            self.var.TransCum = globals.inZero.copy()
        # Cumulative transmission loss
        # self.var.TransLossM3Dt = globals.inZero.copy()
        # substep amount of transmission loss


    def dynamic_inloop(self):
        """ dynamic part of the transmission loss routine
           inside the sub time step routing routine
        """

        # ************************************************************
        # ***** TRANSMISSION LOSS IN THE CHANNEL      ****************
        # ************************************************************
        if option['TransLoss']:


            TransOut = np.where(self.var.UpTrans,
                        (self.var.ChanQ ** self.var.TransPower2 - self.var.TransSub)
                        ** self.var.TransPower1, self.var.ChanQ)
            # transmission loss (equation: Rao and Maurer 1996, Water Resources
            # Bulletin Vol 32, No.6)

            self.var.TransLossM3Dt = (self.var.ChanQ - TransOut) * self.var.DtRouting

            #self.var.TransLossM3Dt = cover((self.var.ChanQ - TransOut) * self.var.DtRouting, scalar(0.0))
            # Loss is Q - transmission outflow
            self.var.TransCum += self.var.TransLossM3Dt
            # for mass balance

